import functools
import logging
from enum import Enum
from pathlib import Path
from typing import Callable, Optional, Sequence

import numpy as np
import pandas as pd
import sklearn.base
import yacs.config as config

from jahs_bench.surrogate import config as surrogate_config
from jahs_bench.surrogate import loss

_log = logging.getLogger(__name__)


def custom_loss_function(loss_params: config.CfgNode) -> Callable:
    """ Given a configuration, construct a custom loss function. """

    func_map = {
        surrogate_config.default_base_loss_configs.exp_lower_bound.name: loss.exponential_bounds,
        surrogate_config.default_base_loss_configs.exp_upper_bound.name: loss.exponential_bounds,
        surrogate_config.default_base_loss_configs.se.name: loss.squared_error
    }

    funcs = []
    weights = []
    for f, p, w in zip(loss_params.funcs, loss_params.params, loss_params.weights):
        if not f:
            continue
        funcs.append(functools.partial(func_map[f], **p))
        weights.append(w)

    loss_func = loss.mix_objectives(*funcs, weights=weights)
    return loss_func


def load_pipeline_config(*params) -> config.CfgNode:
    """ Given a list of parameters, interprets the list as (key, value) pairs and
    converts them into an appropriate pipeline configuration. """

    pipeline_config = surrogate_config.default_pipeline_config.clone()
    pipeline_config.defrost()
    pipeline_config.set_new_allowed(True)
    pipeline_config.merge_from_list(params)

    try:
        loss_type = surrogate_config.RegressionLossFuncTypes(pipeline_config.loss)
    except ValueError as e:
        raise ValueError(
            f"The loss function for the pipeline should be a right-hand string from the "
            f"following mapping:\n{list(surrogate_config.RegressionLossFuncTypes)}") from e

    if loss_type is surrogate_config.RegressionLossFuncTypes.squared_error:
        return pipeline_config

    # Assume it's a custom loss function
    try:
        for f in pipeline_config.loss_params.funcs:
            if not f:
                continue
            _ = surrogate_config.default_base_loss_configs(f)
    except ValueError as e:
        raise ValueError(
            f"Unrecognized base loss function {f}, must be either '' or one of "
            f"{list(surrogate_config.default_base_loss_configs.__members__.keys())}") from e

    loss_params = pipeline_config.loss_params
    for i, p in enumerate(loss_params.params):
        # The parameters may be specified by passing in a filename instead
        if isinstance(p, str):
            cfg_file = Path(pipeline_config.config_dir) / p
            new_params = config.CfgNode(new_allowed=True).merge_from_file(cfg_file)
            loss_params.params[i] = new_params

    target_config = pipeline_config.target_config
    for i, p in enumerate(target_config.params):
        # The parameters may be specified by passing in a filename instead
        if isinstance(p, str):
            cfg_file = Path(pipeline_config.config_dir) / p
            new_params = config.CfgNode(new_allowed=True).merge_from_file(cfg_file)
            target_config.params = new_params

    pipeline_config.set_new_allowed(False)
    pipeline_config.freeze()
    return pipeline_config


class CustomTransformFunctions:
    @staticmethod
    def sigmoid(arr: np.array, k: float = 1.):
        if k <= 0:
            raise ValueError(f"The parameter 'k' for a sigmoid function must be a "
                             f"positive real number, was {k}.")
        return np.power(1 + np.exp(-k * arr), -1)

    @staticmethod
    def inverse_sigmoid(arr: np.array, k: float = 1.):
        if k <= 0:
            raise ValueError(f"The parameter 'k' for an inverse sigmoid function must be "
                             f"a positive real number, was {k}.")
        if not (arr.min() > 0. and arr.max() < 1.0):
            raise ValueError(f"The domain of the inverse sigmoid function is the open "
                             f"interval (0, 1), but the given inputs have span the "
                             f"closed interval [{arr.min()}, {arr.max()}].")
        return (np.log(arr) - np.log(1 - arr)) / k


# Valid transformations for the target variable
class TargetTransforms(Enum):
    MinMax = sklearn.preprocessing.MinMaxScaler


class CustomTransforms(Enum):
    sigmoid = (CustomTransformFunctions.sigmoid, CustomTransformFunctions.inverse_sigmoid)
    inverse_sigmoid = \
        (CustomTransformFunctions.inverse_sigmoid, CustomTransformFunctions.sigmoid)


def apply_target_transform(target_config: config.CfgNode, final_estimator) -> \
    sklearn.base.BaseEstimator:
    """ Given a configuration describing how the targets should be transformed and a
    corresponding estimator, returns a MetaEstimator containing a pipeline such that the
    required transformations are applied to the targets before passed to
    'final_estimator'. The 'target_config' may describe either a single transformation or
    a chain of transformations. """

    if target_config.transform.lower() == "chain":
        # Chain the transformations by "prepending" each transformation (starting from
        # the last one) to the original estimator. This is necessary because sklearn
        # Pipelines don't support target transformations.
        estimator = final_estimator
        funcs = target_config.params["funcs"][::-1]
        params = target_config.params["params"][::-1]
        for f, p in zip(funcs, params):
            transformer = _get_single_transform(f, **p)
            estimator = sklearn.compose.TransformedTargetRegressor(
                transformer=transformer, regressor=estimator)
    else:
        transformer = _get_single_transform(target_config.transform,
                                            **target_config.params)
        estimator = sklearn.compose.TransformedTargetRegressor(
            transformer=transformer, regressor=final_estimator)

    return estimator


def _get_single_transform(name: str, **params) -> sklearn.base.BaseEstimator:
    """ Given the name of a known Transform and the relevant parameters, initialize a
    corresponding Transformer object and return it. """

    if "custom" in name:
        # Construct a custom transformer
        fname = name.split(":")[1]
        func, inv = CustomTransforms[fname].value
        return sklearn.preprocessing.FunctionTransformer(
            func=func, inverse_func=inv, kw_args=params, inv_kw_args=params)
    else:
        # Use an existing Transformer
        return TargetTransforms[name].value(**params)


def adjust_dataset(datapth: Path, outputs: Optional[Sequence[str]] = None,
                   fillna: bool = False):
    _log.info(f"Adjusting data from {datapth}. "
              f"{(' Chosen output columns: ' + str(outputs)) * (outputs is not None)}. "
              f"{'Filling ' if fillna else 'Not filling '} in NaN values. ")
    _log.info("Fetching data.")
    data = pd.read_pickle(datapth)
    _log.info("Finished loading data.")

    _log.info(f"There are {data.index.size} rows of data, including the test set (if "
              f"any).")
    index = data.loc[:, "sampling_index"]
    groups = index["model_ID"]
    strata = index["fidelity_ID"]
    features = data.features
    labels: pd.DataFrame = data.labels

    if outputs is not None:
        labels: pd.DataFrame = labels[outputs]

    if fillna:
        sel: pd.Series = labels.isna().any(axis=1)
        subdf = labels.loc[sel]
        if subdf.shape[0] == 0:
            _log.info("Found no NaN values in the labels.")
        else:
            _log.info(f"Found {subdf.shape[0]} rows of data with NaN values in them. "
                      f"Attempting to fill NaN values.")
            fillvals = {}
            for c in sel.index:
                if sel[c]:
                    if "acc" in c:
                        fillvals[c] = 0.
                    elif "loss" in c or "duration" in c:
                        fillvals[c] = 100 * labels[c].max()
                    else:
                        raise RuntimeError(f"Could not find appropriate fill value to "
                                           f"replace NaNs with for metric: {c}")
            _log.info(f"Filling NaN values according to the mapping: {fillvals}")
            labels.fillna(fillvals, inplace=True)

    return features, labels, groups, strata



import logging
import time
from typing import Optional, Iterable, Callable, Dict, List

import numpy as np
import psutil
import torch
from torch.utils.tensorboard import SummaryWriter

import jahs_bench.tabular.lib.naslib.utils.logging as naslib_logging
import jahs_bench.tabular.lib.naslib.utils.utils as naslib_utils

from jahs_bench.tabular.search_space import NASB201HPOSearchSpace
from jahs_bench.tabular.lib.core import constants, utils
from jahs_bench.tabular.lib.core.custom_nasb201_code import CosineAnnealingLR
from jahs_bench.tabular.lib.core.count_flops import get_model_flops
from jahs_bench.tabular.lib.naslib.utils.utils import AverageMeter, AttrDict


def construct_model_optimizer(model: NASB201HPOSearchSpace, train_config: AttrDict):
    optim_type = utils.Optimizers.__members__[model.config["Optimizer"]]
    optimizer = optim_type.construct(model, model.config)
    if optim_type == utils.Optimizers.SGD:
        # Potentially stabilize SGD with Warm-up
        scheduler = CosineAnnealingLR(optimizer, warmup_epochs=train_config.warmup_epochs, epochs=train_config.sched_max_temp,
                                      T_max=train_config.sched_max_temp, eta_min=0.)
    else:
        # No need to use warmup epochs for Adam and AdamW
        scheduler = CosineAnnealingLR(optimizer, warmup_epochs=0, epochs=train_config.sched_max_temp,
                                      T_max=train_config.sched_max_temp, eta_min=0.)
    loss_fn = torch.nn.CrossEntropyLoss()

    return optimizer, scheduler, loss_fn


def _main_proc(model, dataloader: Iterable, loss_fn: Callable, optimizer: torch.optim.Optimizer,
               scheduler: CosineAnnealingLR, mode: str, device: str, epoch_metrics: Dict[str, List],
               debug: bool = False, use_grad_clipping: bool = True,
               summary_writer: torch.utils.tensorboard.SummaryWriter = None, name: Optional[str] = None, logger=None):
    # Logging setup
    name_str = f"Dataset: {name} " if name else ""
    if not logger:
        logger = logging.getLogger(__name__)

    if mode == "train":
        model.train()
    elif mode == "eval":
        model.eval()
    else:
        raise ValueError(f"Unrecognized mode '{mode}'.")

    ## Setup control flags
    transfer_devices = device.type != "cpu"
    train_model = mode == "train"

    ## Setup metrics
    extra_metrics = []
    # metrics = _get_common_metrics()

    if transfer_devices:
        extra_metrics.append("data_transfer_duration")
        # metrics["data_transfer_time"] = AverageMeter()

    if train_model:
        extra_metrics.append("backprop_duration")
        # metrics["backprop_duration"] = AverageMeter()

    metrics = utils.attrdict_factory(metrics=constants.standard_model_dataset_metrics + extra_metrics)

    if debug:
        losses = []
        accs = []
        grad_norms = []
        lrs = []
        assert isinstance(summary_writer, torch.utils.tensorboard.SummaryWriter), \
            "When using debug mode, a tensorboard summary writer object must be passed."

    ## Iterate over mini-batches
    diverged = False
    data_load_start_time = time.time()
    nsteps = len(dataloader)
    for step, (inputs, labels) in enumerate(dataloader):
        metric_weight = inputs.size(0)
        start_time = time.time()
        metrics.data_load_duration.update(start_time - data_load_start_time, metric_weight)

        if train_model:
            scheduler.update(None, 1. * step / nsteps)
            optimizer.zero_grad()

        if transfer_devices:
            inputs = inputs.to(device)
            labels = labels.to(device)
            metrics.data_transfer_duration.update(time.time() - start_time, metric_weight)

        ## Forward Pass
        forward_start_time = time.time()
        logits = model(inputs)
        loss = loss_fn(logits, labels)
        metrics.forward_duration.update(time.time() - forward_start_time, metric_weight)

        ## Backward Pass
        if train_model:
            backprop_start_time = time.time()
            loss.backward()
            if use_grad_clipping:
                torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=2.0, norm_type=2)
            optimizer.step()
            end_time = time.time()
            metrics.backprop_duration.update(end_time - backprop_start_time, metric_weight)
        else:
            end_time = time.time()

        ## Bookkeeping
        metrics.duration.update(end_time - start_time, metric_weight)
        metrics.loss.update(loss.detach().cpu().data.item(), metric_weight)
        acc = naslib_utils.accuracy(logits.detach().cpu(), labels.detach().cpu(), topk=(1,))[0]
        metrics.acc.update(acc.data.item(), metric_weight)

        ## Debugging, logging and clean-up
        naslib_logging.log_every_n_seconds(
            logging.DEBUG,
            f"[Minibatch {step + 1}/{nsteps}] {name_str}Mode: {mode} Avg. Loss: {metrics.loss.avg:.5f}, "
            f"Avg. Accuracy: {metrics.acc.avg:.5f}, Avg. Time per step: {metrics.duration.avg}",
            n=30, name=logger.name
        )

        if debug:
            losses.append(loss.detach().cpu())
            accs.append(acc.data.item())
            grad_norms.append(torch.norm(
                torch.stack([torch.norm(p.grad.detach(), 2).to(device) for p in model.parameters()]),
                2).detach().cpu())
            lrs.append(scheduler.get_lr()[0])  # Assume that there is only one parameter group

        if torch.isnan(loss):
            logger.debug(f"Encountered NAN loss. Predictions: {logits}\n\nExpected labels: {labels}\n\nLoss: {loss}")
            diverged = True

        if diverged:
            if debug:
                divergent_metrics = {
                    "Minibatch Losses": losses,
                    "Minibatch Accuracies": accs,
                    "Minibatch LRs": lrs,
                    "Minibatch Gradient Norms": grad_norms
                }
                for i in range(step):
                    summary_writer.add_scalars(
                        "divergent_epoch_metrics", {k: v[i] for k, v in divergent_metrics.items()}, i)
                summary_writer.flush()
            return 0, True

        data_load_start_time = time.time()

    n = 0  # Number of individual data points processed
    for key, value in metrics.items():
        epoch_metrics[key].append(value.avg)  # Metrics are recorded as averages per data-point for each epoch
        n = max([n, value.cnt])  # value.cnt is expected to be constant

    return n, False


def train(model: NASB201HPOSearchSpace, data_loaders, train_config: AttrDict, dir_tree: utils.DirectoryTree,
          logger: logging.Logger, validate=False, transfer_devices: bool = False,
          device: [torch.DeviceObjType, str] = "cpu", debug: bool = False, min_shape = None):
    """
    Train the given model using the given data loaders and training configuration. Returns a dict containing
    various metrics.

    Parameters
    ----------
    TODO: Update docs, model must be of type NASB201HPOSearchSpace since its config is being read. Alternatively, the
     required config values can be generalised to function arguments. Also, the model should be parsed BEFORE being
     passed.
    model: naslib.search_spaces.Graph
        A NASLib object obtained by sampling a random architecture on a search space. Ideally, the sampled object should
        be cloned before being passsed. The model will be parsed to PyTorch before training.
    data_loaders: Tuple
        A tuple of objects that correspond to the data loaders generated by naslib.utils.utils.get_train_val_loaders().
    min_shape: Tuple[int, int int]
        A 4-tuple of ints corresponding to the smallest valid shape of an input tensor for the given model and a
        single item of the dataset. Can be omitted if debug is False.
    train_config: naslib.utils.utils.AttrDict
        An attribute dict containing various configuration parameters for the model training.
    """

    start_time = time.time()
    latency = AverageMeter()

    # Initialization calls to psutil
    _ = psutil.cpu_percent()
    _ = psutil.virtual_memory()
    _ = psutil.swap_memory()

    train_queue = data_loaders["train"]
    test_queue = data_loaders["test"]
    if validate:
        valid_queue = data_loaders["valid"]

    extra_metrics = ["data_transfer_duration"] if transfer_devices else []
    dataset_metrics = constants.standard_model_dataset_metrics + extra_metrics
    model_metrics = AttrDict({
        "train": utils.attrdict_factory(metrics=dataset_metrics + constants.extra_model_training_metrics,
                                        template=list),
        "valid": utils.attrdict_factory(metrics=dataset_metrics, template=list),
        "test": utils.attrdict_factory(metrics=dataset_metrics, template=list),
        "diagnostic": utils.attrdict_factory(metrics=constants.standard_model_diagnostic_metrics, template=list),
    })

    optimizer, scheduler, loss_fn = construct_model_optimizer(model, train_config)
    if transfer_devices:
        # TODO: Include transfer of model to GPU right here instead of doing it repeatedly in '_main_proc()'.
        loss_fn = loss_fn.to(device)
    logger.debug(f"Initialized optimizer: {optimizer.__class__.__name__}")

    ## Initialize checkpoint function and metric logger, load existing checkpoints and metrics
    old_chkpt_runtime = 0.
    old_chkpt_epoch_idx = -1
    timer = utils.SynchroTimer(ping_interval_time=train_config.checkpoint_interval_seconds,
                               ping_interval_epochs=train_config.checkpoint_interval_epochs)

    if train_config.checkpoint_interval_seconds is not None and train_config.checkpoint_interval_epochs is not None:
        logger.log(logging.WARNING, "Checkpoint logging at regular intervals of time as well as epochs has been "
                                    "enabled simultaneously. This will cause the checkpoints to be saved at either "
                                    "regular time intervals or epoch intervals, depending on which is shorter.")

    if not train_config.disable_checkpointing:
        checkpoint = utils.Checkpointer(model=model, optimizer=optimizer, scheduler=scheduler, dir_tree=dir_tree,
                                        logger=logger, map_location=device, timer=timer)
        old_chkpt_runtime = checkpoint.runtime
        old_chkpt_epoch_idx = checkpoint.last_epoch

    model_metrics_logger = utils.MetricLogger(dir_tree=dir_tree, metrics=model_metrics,
                                              set_type=utils.MetricLogger.MetricSet.model, logger=logger, timer=timer)

    if old_chkpt_runtime != model_metrics_logger.elapsed_runtime:
        raise RuntimeError(f"The Checkpoint and Metrics Log are out of sync. The latest checkpoint has the timestamp "
                           f"{old_chkpt_runtime} while the latest metrics log has the timestamp "
                           f"{model_metrics_logger.elapsed_runtime}. Use the data integrity verification tools to "
                           f"resolve this. Note that this may cause some loss of data.")

    timer.adjust(previous_timestamp=old_chkpt_runtime, last_epoch=old_chkpt_epoch_idx)

    flops = get_model_flops(model=model, input_shape=min_shape, transfer_devices=transfer_devices, device=device)

    if debug:
        summary_writer = SummaryWriter(dir_tree.model_tensorboard_dir)
        model.eval()
        with torch.no_grad():
            summary_writer.add_graph(model, torch.tensor(np.random.random(min_shape).astype(np.float32)).to(device))
    else:
        summary_writer = None

    train_size = valid_size = test_size = 0

    diverged = False

    for e in range(old_chkpt_epoch_idx + 1, train_config.epochs):
        scheduler.update(e, 0.0)
        ## Handle training set
        dataloader = train_queue
        epoch_metrics = model_metrics.train
        n, diverged = _main_proc(model=model, dataloader=dataloader, loss_fn=loss_fn, optimizer=optimizer,
                                 scheduler=scheduler, mode="train", device=device, epoch_metrics=epoch_metrics,
                                 use_grad_clipping=train_config.use_grad_clipping, summary_writer=summary_writer,
                                 name="Training", logger=logger, debug=debug)
        if diverged:
            break
        train_size = max([train_size, n])

        ## Handle validation set, if needed
        if validate:
            dataloader = valid_queue
            epoch_metrics = model_metrics.valid
            with torch.no_grad():
                n, _ = _main_proc(model=model, dataloader=dataloader, loss_fn=loss_fn, optimizer=optimizer,
                                  scheduler=scheduler, mode="eval", device=device, epoch_metrics=epoch_metrics,
                                  use_grad_clipping=train_config.use_grad_clipping, summary_writer=summary_writer,
                                  name="Validation", logger=logger, debug=debug)
            valid_size = max([valid_size, n])
            latency.update(model_metrics.valid.forward_duration[-1], valid_size)
            # Note: eval_duration DOES require a weight - it mixes validation and test sets if validate=True

        ## Handle test set
        dataloader = test_queue
        epoch_metrics = model_metrics.test
        with torch.no_grad():
            n, _ = _main_proc(model=model, dataloader=dataloader, loss_fn=loss_fn, optimizer=optimizer,
                              scheduler=scheduler, mode="eval", device=device, epoch_metrics=epoch_metrics,
                              use_grad_clipping=train_config.use_grad_clipping, summary_writer=summary_writer,
                              name="Test", logger=logger, debug=debug)
        test_size = max([test_size, n])
        latency.update(model_metrics.test.forward_duration[-1], test_size)

        ## Logging
        naslib_logging.log_every_n_seconds(
            logging.DEBUG,
            f"[Epoch {e + 1}/{train_config.epochs}] " +
            f"Avg. Train Loss: {model_metrics.train.loss[-1]:.5f}, " +
            f"Avg. Train Acc: {model_metrics.train.acc[-1]:.5f},\n" +
            (f"Avg. Valid Loss: {model_metrics.valid.loss[-1]:.5f}, " +
             f"Avg. Valid Acc: {model_metrics.valid.acc[-1]:.5f},\n" if validate else "") +
            f"Avg. Test Loss: {model_metrics.test.loss[-1]:.5f}, " +
            f"Avg. Test Acc: {model_metrics.test.acc[-1]:.5f}\n" +
            f"Time Elapsed: {time.time() - start_time}",
            15,
            name=logger.name
        )

        ## Checkpointing
        # Add a one-time offset to the runtime in case an old checkpoint was loaded
        effective_elapsed_runtime = time.time() - start_time + old_chkpt_runtime
        first_epoch = e == 0
        last_epoch = e == train_config.epochs - 1
        timer.update(timestamp=effective_elapsed_runtime, curr_epoch=e, force=first_epoch or last_epoch)
        if not train_config.disable_checkpointing:
            checkpoint()

        # No need to calculate this again everytime, changes reflect checkpointing-relevant differences.
        model_metrics.diagnostic.FLOPS.append(flops)

        model_metrics.diagnostic.latency.append(latency.avg)
        model_metrics.diagnostic.runtime.append(effective_elapsed_runtime)
        model_metrics.diagnostic.cpu_percent.append(psutil.cpu_percent())
        model_metrics.diagnostic.memory_ram.append(psutil.virtual_memory().available)
        model_metrics.diagnostic.memory_swap.append(psutil.swap_memory().free)
        model_metrics_logger.log()

    if debug:
        tb_metrics = AttrDict(
            train_loss=model_metrics.train.loss,
            train_acc=model_metrics.train.acc,
            test_loss=model_metrics.test.loss,
            test_acc=model_metrics.test.acc,
            **(dict(valid_loss=model_metrics.valid.loss, valid_acc=model_metrics.valid.acc, ) if validate else {})
        )
        # If training diverged, the last epoch was not completed and should be looked at in detail instead.
        num_epochs = train_config.epochs if not diverged else e - 1
        if num_epochs >= 0:
            for i in range(num_epochs):
                summary_writer.add_scalars("metrics", {k: v[i] for k, v in tb_metrics.items()})

        summary_writer.flush()
        summary_writer.close()

    if diverged:
        raise RuntimeError(f"Model training has diverged after {e} epochs.")

    if not train_config.disable_checkpointing:
        del checkpoint
    del model_metrics_logger

    return model_metrics



import jahs_bench


def run():
    benchmark = jahs_bench.Benchmark(task="cifar10", kind="surrogate", download=True)
    config = benchmark.sample_config()
    results = benchmark(config, nepochs=200)
    print(f"Sampled random configuration: {config}\nResults of query on the surrogate "
          f"model: {results}")


if __name__ == "__main__":
    run()



import ConfigSpace as CS
from .constants import Activations

joint_config_space = CS.ConfigurationSpace("NASB201HPO_joint_config_space")
joint_config_space.add_hyperparameters([
    CS.OrdinalHyperparameter("N", sequence=[1, 3, 5], default_value=1, meta=dict(help="Number of cell repetitions")),
    # CS.OrdinalHyperparameter("W", sequence=list(range(1, 4)), default_value=1,
    #                          meta=dict(help="The width of the three channels in the cell. The value of this "
    #                                         "parameter corresponds to one of three levels:\n1 - (4, 8, 16)\n"
    #                                         "2 - (8, 16, 32)\n3 - (16, 32, 64)")),
    CS.OrdinalHyperparameter("W", sequence=[4, 8, 16], default_value=4,
                             meta=dict(help="The width of the first channel in the cell. Each of the subsequent "
                                            "cell's first channels is twice as wide as the previous cell's, thus, for "
                                            "a value 4 (default) of W, the first channel widths are [4, 8, 16].")),
    CS.CategoricalHyperparameter("Op1", choices=list(range(5)), default_value=0,
                                 meta=dict(help="The operation on the first edge of the cell.")),
    CS.CategoricalHyperparameter("Op2", choices=list(range(5)), default_value=0,
                                 meta=dict(help="The operation on the second edge of the cell.")),
    CS.CategoricalHyperparameter("Op3", choices=list(range(5)), default_value=0,
                                 meta=dict(help="The operation on the third edge of the cell.")),
    CS.CategoricalHyperparameter("Op4", choices=list(range(5)), default_value=0,
                                 meta=dict(help="The operation on the fourth edge of the cell.")),
    CS.CategoricalHyperparameter("Op5", choices=list(range(5)), default_value=0,
                                 meta=dict(help="The operation on the fifth edge of the cell.")),
    CS.CategoricalHyperparameter("Op6", choices=list(range(5)), default_value=0,
                                 meta=dict(help="The operation on the sixth edge of the cell.")),
    CS.OrdinalHyperparameter("Resolution", sequence=[0.25, 0.5, 1.], default_value=1.,
                             meta=dict(help="The sample resolution of the input images w.r.t. one side of the actual "
                                            "image size, assuming square images, i.e. for a dataset with 32x32 images, "
                                            "specifying a value of 0.5 corresponds to using downscaled images of size "
                                            "16x16 as inputs.")),
    CS.CategoricalHyperparameter("TrivialAugment", choices=[True, False], default_value=False,
                                 meta=dict(help="Controls whether or not TrivialAugment is used for pre-processing "
                                                "data. If False (default), a set of manually chosen transforms is "
                                                "applied during pre-processing. If True, these are skipped in favor "
                                                "of applying random transforms selected by TrivialAugment.")),
    CS.CategoricalHyperparameter("Activation", choices=list(Activations.__members__.keys()),
                                 default_value="ReLU", meta=dict(help="Which activation function is to be used for the "
                                                                      "network. Default is ReLU."))
])

# Add Optimizer related HyperParamters
optimizers = CS.CategoricalHyperparameter("Optimizer", choices=["AdamW", "SGD"], default_value="SGD",
                                          meta=dict(help="Which optimizer to use for training this model. SGD refers "
                                                         "to SGD with Nesterov momentum and a momentum of 0.9. AdamW "
                                                         "refers to AdamW with betas fixed to 0.9 and 0.999."))
lr = CS.UniformFloatHyperparameter("LearningRate", lower=1e-3, upper=1e0, default_value=1e-1, log=True,
                                   meta=dict(help="The learning rate for the optimizer used during model training. "
                                                  "In the case of adaptive learning rate optimizers such as Adam, "
                                                  "this is the initial learning rate."))
weight_decay = CS.UniformFloatHyperparameter("WeightDecay", lower=1e-5, upper=1e-2, default_value=5e-4, log=True,
                                             meta=dict(help="Weight decay to be used by the optimizer during model "
                                                            "training."))

joint_config_space.add_hyperparameters([optimizers, lr, weight_decay])



""" A collection of constant values that are or may be re-used throughout the rest of the code base. """

from enum import Enum

training_config = {
    "epochs": dict(type=int, default=25,
                   help="Number of epochs that each sampled architecture should be trained for. Default: 25"),
    "batch_size": dict(type=int, default=256, help="Number of samples per mini-batch."),
    "use_grad_clipping": dict(action="store_true", help="Enable gradient clipping for SGD."),
    "split": dict(action="store_true", help="Split training dataset into training and validation sets."),
    "warmup_epochs": dict(type=int, default=0,
                          help="When set to a positive integer, this many epochs are used to warm-start the training."),
    "sched_max_temp": dict(type=int, default=200,
                   help="Property of the learning rate scheduler, determines the maximum number of epochs that the "
                        "scheduler decays the learning rate over. Default: 200"),
    "disable_checkpointing": dict(
        action="store_true",
        help="When given, overrides the values of --checkpoint_interval_seconds and --checkpoint_internal_epochs "
             "and disables checkpointing of model weights. Normally, model weights and meitrcs are checkpointed "
             "either every X epochs or Y seconds. Check --checkpoint_interval_seconds and --checkpoint_interval_epochs "
             "for more information."),
    "checkpoint_interval_seconds": dict(
        type=int, default=None,
        help="The time interval between subsequent model training checkpoints and metrics, in seconds. When not "
             "specified (default), model training checkpointing at regular intervals of time is turned off. When "
             "specified along with --checkpoint_interval_epochs, both checkpointing intervals are enabled "
             "simultaneously, i.e. the model training will be logged either every X epochs and Y seconds, depending on "
             "which condition is satisfied first. On the flip side, if neither --checkpoint_interval_epochs nor "
             "--checkpoint_interval_seconds is specified, logging of model training is disabled entirely. Even if "
             "either or both these values are specified, checkpointing of model weights can be forcefully disabled by "
             "the flag --disable_checkpointing."),
    "checkpoint_interval_epochs": dict(
        type=int, default=None,
        help="The interval between subsequent model training checkpoints and metrics, in epochs. When not specified "
             "(default), model training checkpointing at a regular number of epochs is turned off. When specified "
             "along with --checkpoint_interval_seconds, both checkpointing intervals are enabled simultaneously, i.e. "
             "the model training will be logged either every X epochs and Y seconds, depending on which condition is "
             "satisfied first. On the flip side, if neither --checkpoint_interval_epochs nor "
             "--checkpoint_interval_seconds is specified, logging of model training is disabled entirely. Even if "
             "either or both these values are specified, checkpointing of model weights can be forcefully disabled by "
             "the flag --disable_checkpointing."),
}


class Datasets(Enum):
    """ Constants corresponding to datasets, each value is a 4 tuple corresponding to the format:
        name: str,
        image_size: int,
        nchannels: int,
        nclasses: int,
        mean: tuple(int, int, int) or tuple(int),
        std: tuple(int, int, int) or tuple(int),
        expected_train_size: int,
        expected_test_size: int,
    """

    cifar10 = "Cifar-10", 32, 3, 10, [0.4914, 0.4822, 0.4465], [0.2470, 0.2435, 0.2616], 50_000, 10_000

    # For ICGen datasets, the mean and std will be read from the accompanying JSON file and not stored statically.
    # This is an ugly hack, but was necessary at the time. In future iterations, either Cifar10/fashionMNIST will also
    # be read as ICGen datasets or a different approach will be adopted than this Enum.
    uc_merced = "UC-Merced", 32, 3, 21, [0., 0., 0.], [1., 1., 1.], 1_890, 210
    colorectal_histology = "Colorectal Histology", 32, 3, 8, [0., 0., 0.], [1., 1., 1.], 4_504, 496
    fashion_mnist = "FashionMNIST", 32, 1, 10, [0., ], [0., ], 63_000, 7_000


icgen_datasets = (Datasets.uc_merced, Datasets.colorectal_histology, Datasets.fashion_mnist)

## Metric DataFrame constant
class MetricDFIndexLevels(Enum):
    # taskid = "TaskIndex" # intended new, will require lots of modifications to AttrDict member accesses
    # modelid = "ModelIndex" # intended new
    taskid = "taskid" # old, to be replaced
    modelid = "model_idx" # old, to be replaced
    epoch = "Epoch"

metricdf_column_level_names = ["MetricType", "MetricName"]
metricdf_index_levels = list(MetricDFIndexLevels.__members__.keys())

## Metric metadata constants
standard_task_metrics = [MetricDFIndexLevels.modelid.value, "model_config", "global_seed", "size_MB"]
standard_model_dataset_metrics = ["duration", "data_load_duration", "forward_duration", "loss", "acc"]
extra_model_training_metrics = ["backprop_duration"]
standard_model_diagnostic_metrics = ["FLOPS", "latency", "runtime", "cpu_percent", "memory_ram", "memory_swap"]

fidelity_types = {"N": int, "W": int, "Resolution": float}
fidelity_params = tuple(fidelity_types.keys())
# TODO: Move more constants here so as to save the effort of re-typing and re-checking, say, string values, repeatedly.



""" This module and its sub-modules have been adapted from or taken as-is from and with
the permission of the authors/maintainers of NASLib, at
https://github.com/automl/NASLib. """

import logging

_log = logging.getLogger(__name__)



""" This file has been used as-is and with permission from the authors/maintainers of
NASLib, at https://github.com/automl/NASLib."""

import logging
import sys
import functools
import os
import torch
import time

from collections import Counter
from termcolor import colored


def log_formats(x):
    if isinstance(x, torch.Tensor):
        return x.shape
    if isinstance(x, dict):
        return {k: log_formats(v) for k, v in x.items()}
    else:
        return x


"""
taken from FAIR's detectron2
"""


class _ColorfulFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        self._root_name = kwargs.pop("root_name") + "."
        self._abbrev_name = kwargs.pop("abbrev_name", "")
        if len(self._abbrev_name):
            self._abbrev_name = self._abbrev_name + "."
        super(_ColorfulFormatter, self).__init__(*args, **kwargs)

    def formatMessage(self, record):
        record.name = record.name.replace(self._root_name, self._abbrev_name)
        log = super(_ColorfulFormatter, self).formatMessage(record)
        if record.levelno == logging.WARNING:
            prefix = colored("WARNING", "red", attrs=["blink"])
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            prefix = colored("ERROR", "red", attrs=["blink", "underline"])
        else:
            return log
        return prefix + " " + log


@functools.lru_cache()  # so that calling setup_logger multiple times won't add many handlers
def setup_logger(output=None, *, color=True, name="naslib", abbrev_name=None):
    """
    Initialize the nalsib logger and set its verbosity level to "DEBUG".
    Args:
        output (str): a file name or a directory to save log. If None, will not save log file.
            If ends with ".txt" or ".log", assumed to be a file name.
            Otherwise, logs will be saved to `output/log.txt`.
        name (str): the root module name of this logger
        abbrev_name (str): an abbreviation of the module, to avoid long names in logs.
            Set to "" to not log the root module in logs.
            By default, will abbreviate "naslib" to "nl" and leave other
            modules unchanged.
    Returns:
        logging.Logger: a logger
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    if abbrev_name is None:
        abbrev_name = "nl" if name == "naslib" else name

    plain_formatter = logging.Formatter(
        "[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S"
    )

    # stdout logger
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.INFO)
    if color:
        formatter = _ColorfulFormatter(
            colored("[%(asctime)s %(name)s]: ", "green") + "%(message)s",
            datefmt="%m/%d %H:%M:%S",
            root_name=name,
            abbrev_name=str(abbrev_name),
        )
    else:
        formatter = plain_formatter
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # File logging
    if output is not None:
        if output.endswith(".txt") or output.endswith(".log"):
            filename = output
        else:
            filename = os.path.join(output, "log.txt")

        if os.path.dirname(filename):
            os.makedirs(os.path.dirname(filename), exist_ok=True)

        fh = logging.FileHandler(filename)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(plain_formatter)
        logger.addHandler(fh)

    return logger


"""
Below are some other convenient logging methods.
They are mainly adopted from
https://github.com/abseil/abseil-py/blob/master/absl/logging/__init__.py
"""


def _find_caller():
    """
    Returns:
        str: module name of the caller
        tuple: a hashable key to be used to identify different callers
    """
    frame = sys._getframe(2)
    while frame:
        code = frame.f_code
        if os.path.join("utils", "logger.") not in code.co_filename:
            mod_name = frame.f_globals["__name__"]
            if mod_name == "__main__":
                mod_name = "detectron2"
            return mod_name, (code.co_filename, frame.f_lineno, code.co_name)
        frame = frame.f_back


_LOG_COUNTER = Counter()
_LOG_TIMER = {}


def log_first_n(lvl, msg, n=1, *, name=None, key="caller"):
    """
    Log only for the first n times.
    Args:
        lvl (int): the logging level
        msg (str):
        n (int):
        name (str): name of the logger to use. Will use the caller's module by default.
        key (str or tuple[str]): the string(s) can be one of "caller" or
            "message", which defines how to identify duplicated logs.
            For example, if called with `n=1, key="caller"`, this function
            will only log the first call from the same caller, regardless of
            the message content.
            If called with `n=1, key="message"`, this function will log the
            same content only once, even if they are called from different places.
            If called with `n=1, key=("caller", "message")`, this function
            will not log only if the same caller has logged the same message before.
    """
    if isinstance(key, str):
        key = (key,)
    assert len(key) > 0

    caller_module, caller_key = _find_caller()
    hash_key = ()
    if "caller" in key:
        hash_key = hash_key + caller_key
    if "message" in key:
        hash_key = hash_key + (msg,)

    _LOG_COUNTER[hash_key] += 1
    if _LOG_COUNTER[hash_key] <= n:
        logging.getLogger(name or caller_module).log(lvl, msg)


def log_every_n(lvl, msg, n=1, *, name=None):
    """
    Log once per n times.
    Args:
        lvl (int): the logging level
        msg (str):
        n (int):
        name (str): name of the logger to use. Will use the caller's module by default.
    """
    caller_module, key = _find_caller()
    _LOG_COUNTER[key] += 1
    if n == 1 or _LOG_COUNTER[key] % n == 1:
        logging.getLogger(name or caller_module).log(lvl, msg)


def log_every_n_seconds(lvl, msg, n=1, *, name=None):
    """
    Log no more than once per n seconds.
    Args:
        lvl (int): the logging level
        msg (str):
        n (int):
        name (str): name of the logger to use. Will use the caller's module by default.
    """
    caller_module, key = _find_caller()
    last_logged = _LOG_TIMER.get(key, None)
    current_time = time.time()
    if last_logged is None or current_time - last_logged >= n:
        logging.getLogger(name or caller_module).log(lvl, msg)
        _LOG_TIMER[key] = current_time



import torch
import torch.nn as nn
from torch.autograd import Variable

from abc import ABCMeta, abstractmethod


class AbstractPrimitive(nn.Module, metaclass=ABCMeta):
    """
    Use this class when creating new operations for edges.

    This is required because we are agnostic to operations
    at the edges. As a consequence, they can contain subgraphs
    which requires naslib to detect and properly process them.
    """

    def __init__(self, kwargs):
        super().__init__()

        self.init_params = {
            k: v
            for k, v in kwargs.items()
            if k != "self" and not k.startswith("_") and k != "kwargs"
        }

    @abstractmethod
    def forward(self, x, edge_data):
        """
        The forward processing of the operation.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_embedded_ops(self):
        """
        Return any embedded ops so that they can be
        analysed whether they contain a child graph, e.g.
        a 'motif' in the hierachical search space.

        If there are no embedded ops, then simply return
        `None`. Should return a list otherwise.
        """
        raise NotImplementedError()

    @property
    def get_op_name(self):
        return type(self).__name__


class Identity(AbstractPrimitive):
    """
    An implementation of the Identity operation.
    """

    def __init__(self, **kwargs):
        super().__init__(locals())

    def forward(self, x, edge_data=None):
        return x

    def get_embedded_ops(self):
        return None


class Zero(AbstractPrimitive):
    """
    Implementation of the zero operation. It removes
    the connection by multiplying its input with zero.
    """

    def __init__(self, stride, **kwargs):
        """
        When setting stride > 1 then it is assumed that the
        channels must be doubled.
        """
        super().__init__(locals())
        self.stride = stride

    def forward(self, x, edge_data=None):
        if self.stride == 1:
            return x.mul(0.0)
        else:
            return x[:, :, :: self.stride, :: self.stride].mul(0.0)

    def get_embedded_ops(self):
        return None

    def __repr__(self):
        return "Zero (stride={})".format(self.stride)


class Zero1x1(AbstractPrimitive):
    """
    Implementation of the zero operation. It removes
    the connection by multiplying its input with zero.
    """

    def __init__(self, stride, **kwargs):
        """
        When setting stride > 1 then it is assumed that the
        channels must be doubled.
        """
        super().__init__(locals())
        self.stride = stride

    def forward(self, x, edge_data):
        if self.stride == 1:
            return x.mul(0.0)
        else:
            x = x[:, :, :: self.stride, :: self.stride].mul(0.0)
            return torch.cat([x, x], dim=1)  # double the channels TODO: ugly as hell

    def get_embedded_ops(self):
        return None

    def __repr__(self):
        return "Zero1x1 (stride={})".format(self.stride)


class SepConv(AbstractPrimitive):
    """
    Implementation of Separable convolution operation as
    in the DARTS paper, i.e. 2 sepconv directly after another.
    """

    def __init__(
        self, C_in, C_out, kernel_size, stride, padding, affine=True, **kwargs
    ):
        super().__init__(locals())
        self.kernel_size = kernel_size
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(
                C_in,
                C_in,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                groups=C_in,
                bias=False,
            ),
            nn.Conv2d(C_in, C_in, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_in, affine=affine),
            nn.ReLU(inplace=False),
            nn.Conv2d(
                C_in,
                C_in,
                kernel_size=kernel_size,
                stride=1,
                padding=padding,
                groups=C_in,
                bias=False,
            ),
            nn.Conv2d(C_in, C_out, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_out, affine=affine),
        )

    def forward(self, x, edge_data=None):
        return self.op(x)

    def get_embedded_ops(self):
        return None

    @property
    def get_op_name(self):
        op_name = super().get_op_name
        op_name += "{}x{}".format(self.kernel_size, self.kernel_size)
        return op_name


class DilConv(AbstractPrimitive):
    """
    Implementation of a dilated separable convolution as
    used in the DARTS paper.
    """

    def __init__(
        self, C_in, C_out, kernel_size, stride, padding, dilation, affine=True, **kwargs
    ):
        super().__init__(locals())
        self.kernel_size = kernel_size
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(
                C_in,
                C_in,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                groups=C_in,
                bias=False,
            ),
            nn.Conv2d(C_in, C_out, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(C_out, affine=affine),
        )

    def forward(self, x, edge_data):
        return self.op(x)

    def get_embedded_ops(self):
        return None

    @property
    def get_op_name(self):
        op_name = super().get_op_name
        op_name += "{}x{}".format(self.kernel_size, self.kernel_size)
        return op_name


class Stem(AbstractPrimitive):
    """
    This is used as an initial layer directly after the
    image input.
    """

    def __init__(self, C_out, **kwargs):
        super().__init__(locals())
        self.seq = nn.Sequential(
            nn.Conv2d(3, C_out, 3, padding=1, bias=False), nn.BatchNorm2d(C_out)
        )

    def forward(self, x, edge_data=None):
        return self.seq(x)

    def get_embedded_ops(self):
        return None


class Sequential(AbstractPrimitive):
    """
    Implementation of `torch.nn.Sequential` to be used
    as op on edges.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(locals())
        self.primitives = args
        self.op = nn.Sequential(*args)

    def forward(self, x, edge_data):
        return self.op(x)

    def get_embedded_ops(self):
        return list(self.primitives)


class MaxPool(AbstractPrimitive):
    def __init__(self, C_in, kernel_size, stride, use_bn=True, **kwargs):
        super().__init__(locals())
        self.kernel_size = kernel_size

        if use_bn:
            self.maxpool = nn.Sequential(
                nn.MaxPool2d(kernel_size, stride=stride, padding=1),
                nn.BatchNorm2d(C_in, affine=False),
            )
        else:
            self.maxpool = nn.MaxPool2d(kernel_size, stride=stride, padding=1)

    def forward(self, x, edge_data):
        x = self.maxpool(x)
        return x

    def get_embedded_ops(self):
        return None


class MaxPool1x1(AbstractPrimitive):
    """
    Implementation of MaxPool with an optional 1x1 convolution
    in case stride > 1. The 1x1 convolution is required to increase
    the number of channels.
    """

    def __init__(
        self, kernel_size, stride, C_in=None, C_out=None, affine=True, **kwargs
    ):
        super().__init__(locals())
        self.stride = stride
        self.maxpool = nn.MaxPool2d(kernel_size, stride=stride, padding=1)
        if stride > 1:
            assert C_in is not None and C_out is not None
            self.conv = nn.Conv2d(C_in, C_out, 1, stride=1, padding=0, bias=False)
            self.bn = nn.BatchNorm2d(C_out, affine=affine)

    def forward(self, x, edge_data):
        x = self.maxpool(x)
        if self.stride > 1:
            x = self.conv(x)
            x = self.bn(x)
        return x

    def get_embedded_ops(self):
        return None


class AvgPool(AbstractPrimitive):
    """
    Implementation of Avergae Pooling.
    """

    def __init__(self, C_in, kernel_size, stride, use_bn=True, **kwargs):
        super().__init__(locals())

        if use_bn:
            self.avgpool = nn.Sequential(
                nn.AvgPool2d(3, stride=stride, padding=1, count_include_pad=False),
                nn.BatchNorm2d(C_in, affine=False),
            )
        else:
            self.avgpool = nn.AvgPool2d(3, stride=stride, padding=1, count_include_pad=False)

    def forward(self, x, edge_data):
        x = self.avgpool(x)
        return x

    def get_embedded_ops(self):
        return None


class AvgPool1x1(AbstractPrimitive):
    """
    Implementation of Avergae Pooling with an optional
    1x1 convolution afterwards. The convolution is required
    to increase the number of channels if stride > 1.
    """

    def __init__(
        self, kernel_size, stride, C_in=None, C_out=None, affine=True, **kwargs
    ):
        super().__init__(locals())
        self.stride = stride
        self.avgpool = nn.AvgPool2d(
            3, stride=stride, padding=1, count_include_pad=False
        )
        if stride > 1:
            assert C_in is not None and C_out is not None
            self.conv = nn.Conv2d(C_in, C_out, 1, stride=1, padding=0, bias=False)
            self.bn = nn.BatchNorm2d(C_out, affine=affine)

    def forward(self, x, edge_data):
        x = self.avgpool(x)
        if self.stride > 1:
            x = self.conv(x)
            x = self.bn(x)
        return x

    def get_embedded_ops(self):
        return None

class GlobalAveragePooling(AbstractPrimitive):
    """
    Just a wrapper class for averaging the input across the height and width dimensions
    """

    def __init__(self):
        super().__init__(locals())

    def forward(self, x, edge_data=None):
        return torch.mean(x, (2, 3))

    def get_embedded_ops(self):
        return None


class ReLUConvBN(AbstractPrimitive):
    """
    Implementation of ReLU activation, followed by 2d convolution and then 2d batch normalization.
    """
    def __init__(self, C_in, C_out, kernel_size, stride=1, affine=True, **kwargs):
        super().__init__(locals())
        self.kernel_size = kernel_size
        pad = 0 if stride == 1 and kernel_size == 1 else 1
        self.op = nn.Sequential(
            nn.ReLU(inplace=False),
            nn.Conv2d(C_in, C_out, kernel_size, stride=stride, padding=pad, bias=False),
            nn.BatchNorm2d(C_out, affine=affine),
        )

    def forward(self, x, edge_data=None):
        return self.op(x)

    def get_embedded_ops(self):
        return None

    @property
    def get_op_name(self):
        op_name = super().get_op_name
        op_name += "{}x{}".format(self.kernel_size, self.kernel_size)
        return op_name

class ConvBnReLU(AbstractPrimitive):
    """
    Implementation of 2d convolution, followed by 2d batch normalization and ReLU activation.
    """
    def __init__(self, C_in, C_out, kernel_size, stride=1, affine=True, **kwargs):
        super().__init__(locals())
        self.kernel_size = kernel_size
        pad = 0 if stride == 1 and kernel_size == 1 else 1
        self.op = nn.Sequential(
            nn.Conv2d(C_in, C_out, kernel_size, stride=stride, padding=pad, bias=False),
            nn.BatchNorm2d(C_out, affine=affine),
            nn.ReLU(inplace=False),
        )

    def forward(self, x, edge_data=None):
        return self.op(x)

    def get_embedded_ops(self):
        return None

    @property
    def get_op_name(self):
        op_name = super().get_op_name
        op_name += "{}x{}".format(self.kernel_size, self.kernel_size)
        return op_name


class InputProjection(AbstractPrimitive):
    """
    Implementation of a 1x1 projection, followed by an abstract primitive model.
    """
    def __init__(self, C_in: int, C_out: int, primitive: AbstractPrimitive):
        """
        Args:
            C_in        : Number of input channels
            C_out       : Number of output channels
            primitive   : Module of AbstractPrimitive type to which the projected input will be fed
        """
        super().__init__(locals())
        self.module = primitive
        self.op = nn.Sequential(
            ConvBnReLU(C_in, C_out, 1), # 1x1 projection
            primitive,                  # Main operation
        )

    def forward(self, x, edge_data):
        return self.op(x)

    def get_embedded_ops(self):
        return None

    @property
    def get_op_name(self):
        op_name = super().get_op_name
        op_name += f"{self.module.get_op_name()}"
        return op_name


class Concat1x1(nn.Module):
    """
    Implementation of the channel-wise concatination followed by a 1x1 convolution
    to retain the channel dimension.
    """

    def __init__(self, num_in_edges, C_out, affine=True, **kwargs):
        super().__init__()
        self.conv = nn.Conv2d(
            num_in_edges * C_out, C_out, kernel_size=1, stride=1, padding=0, bias=False
        )
        self.bn = nn.BatchNorm2d(C_out, affine=affine)

    def forward(self, x):
        """
        Expecting a list of input tensors. Stacking them channel-wise
        and applying 1x1 conv
        """
        x = torch.cat(x, dim=1)
        x = self.conv(x)
        x = self.bn(x)
        return x



from pathlib import Path
import pandas as pd
from string import Template
import math
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--template_dir", type=Path, help="This is the directory where the input templates and "
                                                      "dataframes are stored.")
parser.add_argument("--script_dir", type=Path, help="This is the directory where the generated job scripts will be "
                                                    "stored.")
parser.add_argument("--slurm_dir", type=Path, help="This is the directory where the outputs of the jobs will be "
                                                   "stored i.e. WORK.")
args = parser.parse_args()

pth = args.template_dir
cpus_per_worker_per_node_per_bucket: pd.Series = pd.read_pickle(pth / "cpus_per_worker_per_node_per_bucket.pkl.gz")
cpuh_per_worker_per_bucket: pd.Series = pd.read_pickle(pth / "cpuh_per_worker_per_bucket.pkl.gz")
evals_per_worker: pd.Series = pd.read_pickle(pth / "evals_per_worker.pkl.gz")
nodes_per_bucket: pd.Series = pd.read_pickle(pth / "nodes_per_bucket.pkl.gz")

with open(pth / "job.template") as fp:
    job_template = fp.read()

with open(pth / "config.template") as fp:
    config_template = fp.read()

fids = cpuh_per_worker_per_bucket.index
scriptdir = args.script_dir
scriptdir.mkdir(exist_ok=True, parents=True)

slurm_dir = args.slurm_dir


def timestr(secs: float):
    minutes = math.ceil(secs / 60)
    hours = minutes // 60
    minutes %= 60
    days = hours // 24
    hours %= 24
    return f"{days}-{hours:02}:{minutes:02}"


task_offset = 0
for f in fids:
    ncpus = int(cpus_per_worker_per_node_per_bucket[f])
    cpuh = cpuh_per_worker_per_bucket[f]
    cpuh = timestr(cpuh)
    nsamples = int(evals_per_worker[f])
    nnodes = int(nodes_per_bucket[f])
    ntasks = 48 // ncpus

    job_name = f"{'-'.join(['-'.join([n, str(i)]) for n, i in zip(fids.names, f)])}"
    jobdir = slurm_dir / job_name
    jobdir.mkdir(exist_ok=True)
    (jobdir / "tasks").mkdir(exist_ok=True)
    (jobdir / "logs").mkdir(exist_ok=True)

    job_script = Template(job_template).substitute(cpuh=cpuh, nnodes=nnodes, job_name=job_name, ncpus=ncpus,
                                                   ntasks=ntasks, jobdir=jobdir, scriptdir=scriptdir)
    config_script = Template(config_template).substitute(nsamples=nsamples, task_offset=task_offset, **{n: i for n, i in zip(fids.names, f)})

    with open(scriptdir / f"{job_name}.job", "w") as fp:
        fp.write(job_script)

    with open(scriptdir / f"{job_name}.config", "w") as fp:
        fp.write(config_script)

    task_offset += ntasks * nnodes




import argparse
import logging
from pathlib import Path
import sys

from jahs_bench.tabular.lib.core.constants import fidelity_params
from jahs_bench.tabular.lib.postprocessing import _log as pproc_log
from jahs_bench.tabular.lib.postprocessing import metric_df_ops as metric_ops

_log = logging.getLogger(__name__)

if __name__ == "__main__":
    fmt = logging.Formatter("[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S")
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(fmt)
    _log.addHandler(ch)
    pproc_log.addHandler(ch)

    parser = argparse.ArgumentParser("Collates all results from the DataFrames produced after successful NASLib "
                                     "training runs into a single large DataFrame.")
    parser.add_argument("--basedir", type=Path,
                        help="The base directory for the directory tree, or the root directory within which multiple "
                             "such base directories exist, depending on whether the script is running in 'stand-alone' "
                             "or 'cluster' mode. See '--workerid' for details.")
    parser.add_argument("--outdir", default=None, type=Path,
                        help="The desired output directory where post-processes data will be stored as various "
                             "pickled DataFrame objects. Default: <outdir>/postproc")
    parser.add_argument("--debug", action="store_true", help="When given, enables debug level output.")
    parser.add_argument("--workerid", type=int, default=-1,
                        help="A worker ID for launching this script on a cluster. When a non-negative value is given, "
                             "it is assumed that the script is operating in a cluster with multiple-parallel processes "
                             "and each process will choose a sub-directory within the given 'basedir' as its own base "
                             "directory based on the fidelity parameter values. Passing a negative value (default: -1) "
                             "assumes the script is running in a stand-alone fashion and will directly use the "
                             "provided 'basedir' as the directory tree's base directory.")
    parser.add_argument("--log", type=Path, default=None,
                        help="A log file to which output logs will be saved. Specifying this turns off logging to "
                             "stdout and stderr (some initial logs may still be sent to stdout and stderr).")
    args = parser.parse_args()

    if args.log is not None:
        fh = logging.FileHandler(args.log.resolve())
        fh.setFormatter(fmt)
        _log.removeHandler(ch)
        _log.addHandler(fh)
        pproc_log.removeHandler(ch)
        pproc_log.addHandler(fh)

    if args.debug:
        _log.setLevel(logging.DEBUG)
        pproc_log.setLevel(logging.DEBUG)
    else:
        _log.setLevel(logging.INFO)
        pproc_log.setLevel(logging.WARNING)

    if args.workerid >= 0:
        fids = {"N": [1, 3, 5], "W": [4, 8, 16], "Resolution": [0.25, 0.5, 1.0]}
        wid = args.workerid
        if wid >= 27:
            sys.exit(0)
        ids = [wid % 3, (wid // 3) % 3, (wid // 9) % 3]
        subdir = "-".join([f"{k}-{fids[k][ids[i]]}" for i, k in enumerate(["N", "W", "Resolution"])])

        basedir = args.basedir / subdir
    else:
        basedir = args.basedir

    _log.info(f"Worker {wid}: Beginning metric data postprocessing at {basedir}.")
    try:
        df = metric_ops.load_metric_df(basedir=basedir)
    except Exception as e:
        _log.warning(f"Worker {wid}: Could not load metric DataFrame. Exiting. Cause: {str(e)}")
        sys.exit(1)

    fidelity_confs = [("model_config", f) for f in fidelity_params]
    configs = metric_ops.get_configs(basedir=None, df=df)

    nepochs_200 = metric_ops.get_nepochs(basedir=None, df=df, filter_epochs=200)
    nepochs_all = metric_ops.get_nepochs(basedir=None, df=df, filter_epochs=-1)

    acc_all_epochs = metric_ops.get_accuracies(basedir=None, df=df, include_validation=True)
    # acc_all_epochs = acc_all_epochs.join([configs])
    acc_200epochs = acc_all_epochs.loc[nepochs_200.index]

    loss_all_epochs = metric_ops.get_losses(basedir=None, df=df, include_validation=True)
    # loss_all_epochs = loss_all_epochs.join([configs])
    loss_200epochs = loss_all_epochs.loc[nepochs_200.index]

    nsamples = metric_ops.get_nsamples(basedir=None, df=df, groupby=fidelity_confs, index=fidelity_params)
    runtimes = metric_ops.get_runtimes(basedir=None, df=df, reduce_epochs=True, extra_durations=None)
    # runtimes = runtimes.join([configs])
    remaining_runtimes = metric_ops.estimate_remaining_runtime(basedir=None, df=df, max_epochs=200)
    # remaining_runtimes = remaining_runtimes.join([configs])

    outdir = basedir / "postproc" if args.outdir is None else args.outdir.resolve()
    outdir.mkdir(exist_ok=True, parents=False)

    configs.to_pickle(outdir / "configs.pkl.gz")
    nsamples.to_pickle(outdir / "nsamples.pkl.gz")
    runtimes.to_pickle(outdir / "runtimes.pkl.gz")
    nepochs_all.to_pickle(outdir / "nepochs.pkl.gz")
    remaining_runtimes.to_pickle(outdir / "remaining_runtimes.pkl.gz")
    acc_200epochs.to_pickle(outdir / "acc_200epochs.pkl.gz")
    acc_all_epochs.to_pickle(outdir / "acc_all_epochs.pkl.gz")
    loss_200epochs.to_pickle(outdir / "loss_200epochs.pkl.gz")
    loss_all_epochs.to_pickle(outdir / "loss_all_epochs.pkl.gz")

    _log.info(f"Worker {wid}: Finished postprocessing.")



"""
Collate all the data generated during training into one big pandas DataFrame. This script was designed specifically to
work with the directory structure defined by the class DirectoryTree. The individual DataFrame objects are expected to
contain all their data as columns with the appropriate metrics names. Theoretically, these can be anything, but the
code has been tested for the following for task and model metrics only:

tasks
-----
    - model_idx
    - model_config
    - global_seed
    - size_MB

models - [train/valid/test]
------
    - duration
    - forward_duration
    - data_load_duration
    - loss
    - acc
    - data_transfer_duration
    - backprop_duration

models - diagnostic
    - FLOPS
    - latency
    - runtime
    - cpu_percent
    - memory_ram
    - memory_swap
"""

import argparse
import logging
from pathlib import Path
import sys
from jahs_bench.tabular.lib.postprocessing import collation

_log = logging.getLogger(__name__)

if __name__ == "__main__":
    fmt = logging.Formatter("[%(asctime)s] %(name)s %(levelname)s: %(message)s", datefmt="%m/%d %H:%M:%S")
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(fmt)
    _log.addHandler(ch)
    collation._log.addHandler(ch)

    parser = argparse.ArgumentParser("Collates all results from the DataFrames produced after successful NASLib "
                                     "training runs into a single large DataFrame.")
    parser.add_argument("--basedir", type=Path,
                        help="The base directory for the directory tree, or the root directory within which multiple "
                             "such base directories exist, depending on whether the script is running in 'stand-alone' "
                             "or 'cluster' mode. See '--workerid' for details.")
    parser.add_argument("--file", default=None, type=Path,
                        help="The desired output filename. The filename should not carry any extension as '.pkl.gz' "
                             "will be automatically appended to it, unless the extension is already '.pkl.gz'. "
                             "Default: <basedir>/data.pkl.gz")
    # TODO: Remove the summarize and clenaup flags, this functionality has been moved to data_verification.
    parser.add_argument("--summarize", action="store_true",
                        help="When this flag is given, a summary of the current status of each run is generated and "
                             "saved on a per-task and job-level basis.")
    parser.add_argument("--cleanup", action="store_true",
                        help="When this flag is given, various files such as error descriptions and checkpoints are "
                             "verified for correctness and consistency, and deleted so as to restore the jobs to an "
                             "state.")
    # TODO: Remove anonymize, this functionality is probably useless.
    parser.add_argument("--anonymize", action="store_true",
                        help="When this flag is given, the task and model ids of the metric data are removed.")
    parser.add_argument("--workerid", type=int, default=-1,
                        help="A worker ID for launching this script on a cluster. When a non-negative value is given, "
                             "it is assumed that the script is operating in a cluster with multiple-parallel processes "
                             "and each process will choose a sub-directory within the given 'basedir' as its own base "
                             "directory based on the fidelity parameter values. Passing a negative value (default: -1) "
                             "assumes the script is running in a stand-alone fashion and will directly use the "
                             "provided 'basedir' as the directory tree's base directory.")
    parser.add_argument("--debug", action="store_true", help="When given, enables debug level output.")
    args = parser.parse_args()

    if args.debug:
        _log.setLevel(logging.DEBUG)
        collation._log.setLevel(logging.DEBUG)
    else:
        _log.setLevel(logging.INFO)
        collation._log.setLevel(logging.WARNING)

    if args.workerid >= 0:
        fids = {"N": [1, 3, 5], "W": [4, 8, 16], "Resolution": [0.25, 0.5, 1.0]}
        wid = args.workerid
        if wid >= 27:
            sys.exit(0)
        ids = [wid % 3, (wid // 3) % 3, (wid // 9) % 3]
        subdir = "-".join([f"{k}-{fids[k][ids[i]]}" for i, k in enumerate(["N", "W", "Resolution"])])

        basedir = args.basedir / subdir / "tasks"
        outdir = args.basedir / subdir
    else:
        basedir = args.basedir
        outdir = args.basedir
        wid = 0

    _log.info(f"Worker {wid}: Beginning metric data collation at {basedir}.")

    try:
        collated_df = collation.collate_tasks(basedir=basedir, cleanup=args.cleanup, save_summary=args.summarize,
                                              anonymize=args.anonymize)
    except Exception as e:
        _log.warning(f"Worker {wid}: Could not collate metrics DataFrames. Exiting. Cause: {str(e)}")
        sys.exit(1)

    outfile: Path = outdir / "metrics.pkl.gz" if args.file is None else args.file.resolve()
    if collated_df is None:
        _log.info(f"Worker {wid}: No valid metric data found at: {args.basedir}")
    else:
        collated_df.to_pickle(outfile)

    _log.info(f"Worker {wid}: Finished metric data collation.")



from typing import Dict, Union, Any, Optional, List, Tuple

import ConfigSpace
import pandas as pd


# TODO: Update documentation
def adapt_search_space(
    original_space: Union[ConfigSpace.ConfigurationSpace, Any],
    portfolio: Optional[pd.DataFrame] = None, taskid: Optional[int] = None,
    opts: Optional[Union[Dict[str, Any], List[str]]] = None,
    suffix: Optional[str] = "_custom") -> Union[
        Tuple[ConfigSpace.ConfigurationSpace, bool], bool]:
    """ Given a ConfigurationSpace object and a valid configuration, restricts the respective configuration space
    object and consequently the overall search space by setting the corresponding parameters to constant values. Such
    a configuration may be provided either by means of a portfolio file along with the relevant taskid or a dictionary
    'opts' mapping strings to values of any type, or both in which case values in 'opts' take precedence. In both
    cases, the configuration is read as a dictionary and the function attempts to autonomously restrict the search
    space such that for each key *k*, the corresponding parameter is set to the constant value *v*. An optional
    string 'suffix' may be provided to modify the name of the ConfigSpace object. Default: '_custom'. If None, the
    name remains unchanged. The name also remains unchanged in the case when the given space was not modified at all,
    most likely because no valid keys were present in the configuration dictionary or the dictionary was empty. In
    such a case, False is returned. Otherwise, True is returned. """

    new_consts = {}
    if portfolio is None and opts is None:
        return False
    elif portfolio is not None:
        if taskid is None:
            raise ValueError(
                f"When a portfolio is given, an integer taskid must also be provided. Was given {taskid}")
        else:
            new_consts = portfolio.loc[taskid % portfolio.index.size, :].to_dict()

    # Convert opts from list of string pairs to a mapping from string to string.
    if isinstance(opts, List):
        i = iter(opts)
        opts = {k: v for k, v in zip(i, i)}
    new_consts = {**new_consts, **opts}

    if hasattr(original_space, "config_space"):
        config_space = original_space.config_space
        flag_cs_attr = True
    else:
        config_space = original_space
        flag_cs_attr = False

    known_params = {p.name: p for p in config_space.get_hyperparameters()}

    def param_interpretor(param, value):
        known_config_space_value_types = {
            ConfigSpace.UniformIntegerHyperparameter: int,
            ConfigSpace.UniformFloatHyperparameter: float,
            ConfigSpace.CategoricalHyperparameter: lambda x: type(param.choices[0])(x),
            ConfigSpace.OrdinalHyperparameter: lambda x: type(param.sequence[0])(x),
        }
        return known_config_space_value_types[type(param)](value)

    modified = False

    for arg, val in new_consts.items():
        if arg in known_params:
            old_param = known_params[arg]
            new_val = param_interpretor(old_param, val)
            if isinstance(new_val, bool):
                new_val = int(
                    new_val)  # Because ConfigSpace doesn't allow Boolean constants.
            known_params[arg] = ConfigSpace.Constant(arg, new_val,
                                                     meta=dict(old_param.meta, **dict(
                                                         constant_overwrite=True)))
            modified = True

    if modified:
        new_config_space = ConfigSpace.ConfigurationSpace(
            f"{config_space.name}{suffix if suffix is not None else ''}")
        new_config_space.add_hyperparameters(known_params.values())
        if flag_cs_attr:
            original_space.config_space = new_config_space
        else:
            return new_config_space, modified

    return modified



"""
This script is intended to be used for pre-computing a more efficient schedule for distributing model configuration
evaluations across the workers in a job allocation using data that has been generated from a smaller run over all the
planned samples in the search space.
"""

import argparse
import logging
import pandas as pd
from pathlib import Path
from string import Template
import shutil
import sys
import time

from jahs_bench.tabular.clusterlib import prescheduler as sched_utils
from jahs_bench.tabular.lib.core.utils import AttrDict
from jahs_bench.tabular.distributed_nas_sampling import run_task, get_tranining_config_from_args, _seed
from jahs_bench.tabular.lib.core import constants

from jahs_bench.tabular.lib.naslib.utils.logging import setup_logger


_log = setup_logger(name='tabular_sampling')


def generate_worker_portfolio(workerid: int, nworkers: int, full_profile: pd.DataFrame) -> pd.DataFrame:
    _log.debug(f"Found {full_profile.index.size} configurations to construct portfolio from.")

    if isinstance(full_profile, pd.DataFrame):
        portfolio = full_profile.loc[full_profile.index[workerid::nworkers], :]
    elif isinstance(full_profile, pd.Series):
        portfolio = full_profile[full_profile.index[workerid::nworkers]]
    else:
        raise RuntimeError("Unexpected control flow sequence. This piece of code should be unreachable. Given "
                           f"profile of type {type(full_profile)}.")
    _log.debug(f"Created portfolio of {portfolio.index.size} configs.")
    return portfolio


# noinspection PyShadowingNames,PyProtectedMember
def _handle_debug(args):
    if args.debug:
        _log.setLevel(logging.DEBUG)
    else:
        _log.setLevel(logging.INFO)


# noinspection PyShadowingNames
def _isolate_training_args(args: argparse.Namespace) -> str:
    """ Given parsed CLI arguments, isolates the values of arguments corresponding to the training config and
    re-constructs the corresponding CLI arguments. """
    training_config = vars(args)
    training_args = []
    actions = {
        "store_true": True,
        "store_false": False
    }
    for k, v in constants.training_config.items():
        if k not in training_config or training_config[k] is None:
            # No value given. If this is optional, leave it empty.
            if "action" not in v and "default" not in v:
                # This is not a flag but it also does not have a default value
                raise RuntimeError(f"Cannot parse the given training config parameters into a valid CLI arguments "
                                   f"string, missing value for required parameter {k}")
        else:
            if "action" in v and "store" in v["action"]:
                # This is a flag
                if actions[v["action"]] == training_config[k]:
                    training_args.append(f"--{k}")
                else:
                    continue
            else:
                # This is a key-value pair
                training_args.append(f"--{k}={training_config[k]}")
    return training_args


def default_training_config() -> AttrDict:
    """ Generate a training config using the default arguments. """

    default_parser = argparse.ArgumentParser()
    for k, v in constants.training_config.items():
        default_parser.add_argument(f"--{k}", **v)

    default_args = default_parser.parse_args([])
    return get_tranining_config_from_args(default_args)


# noinspection PyUnusedLocal,PyShadowingNames
def estimate_requirements(args):
    # TODO: Complete and standardize this procedure. For now, use manually generated estimates.
    raise NotImplementedError("This functionality will be properly built and re-structured at a later date.")


# noinspection PyShadowingNames
def init_directory_tree(args):
    _handle_debug(args)
    _log.debug("Starting sub-program: initialize directory tree.")

    ## Estimated time budget requirement
    start = time.time()
    budget = args.budget - 60

    taskid_levelname = constants.MetricDFIndexLevels.taskid.value
    profile_path: Path = args.profile
    workerid = args.workerid_offset + args.workerid

    try:
        profile = pd.read_pickle(profile_path)
        assert profile.index.is_unique, "Each row index in the profile must be unique throughout the profile."
        basedirs: pd.Series = profile.job_config["basedir"]
    except Exception as e:
        raise RuntimeError(f"Failed to read the directory structure from {profile_path}.") from e

    if basedirs is None:
        _log.error(f"Profile {profile_path} contains no data on the basedirs.")
        sys.exit(-1)

    if args.create_tree:
        sched_utils.prepare_directory_structure(
            basedirs=generate_worker_portfolio(workerid=workerid, nworkers=args.nworkers, full_profile=basedirs),
            rootdir=args.rootdir
        )
        return

    _log.info("Pre-sampling model configs.")

    training_config = default_training_config()
    nsamples = basedirs.index.to_frame(index=False).groupby(taskid_levelname).max()
    if isinstance(nsamples, pd.DataFrame):
        nsamples: pd.Series = nsamples[nsamples.columns[0]].rename("nsamples")

    # Only consider this worker's allocation.
    nsamples = generate_worker_portfolio(workerid=workerid, nworkers=args.nworkers, full_profile=nsamples)

    fidelities = profile.model_config
    dataset = constants.Datasets.__members__.get(args.dataset)

    opts = {}
    if args.opts is not None:
        if isinstance(args.opts, list):
            i = iter(args.opts)
            opts = {k: v for k, v in zip(i, i)}
        else:
            raise RuntimeError(f"Unable to parse the search space overrides {args.opts}")

    ## Handle the checkpointing of this worker's portfolio
    worker_chkpt_subdir = "worker_chkpts"
    chkpt_name = f"preschedule_init_worker.pkl.gz"
    assert args.rootdir.exists(), f"The specified root directory {args.rootdir} does not exist."
    worker_chkpt_dir: Path = args.rootdir / worker_chkpt_subdir / str(workerid)
    worker_chkpt_dir.mkdir(exist_ok=True, parents=True)
    chkpt_filepath = worker_chkpt_dir / chkpt_name

    done = None
    if chkpt_filepath.exists():
        try:
            done = pd.read_pickle(chkpt_filepath)
            nsamples = nsamples[done == False]
            fidelities = fidelities.loc[nsamples.index, :]
        except Exception as e:
            _log.info(f"Ran into an error while trying to read {chkpt_filepath}: {str(e)}")
            done = None

    if done is None:
        done = pd.Series(False, index=nsamples.index)
        done.to_pickle(chkpt_filepath)

    end = time.time()
    duration = end - start
    _log.info(f"Worker {workerid}: Setting up the directory tree consumed {duration} seconds.")
    budget -= duration
    start = end
    max_time = 0.

    ntasks = nsamples.index.size
    for i, taskid in enumerate(nsamples.index, start=1):
        if budget < 1.2 * max_time:
            break

        fidelity = fidelities.xs(taskid, level=taskid_levelname).iloc[0]
        basedir = args.rootdir / basedirs.xs(taskid, level=taskid_levelname).iloc[0]
        run_task(basedir=basedir, taskid=taskid,
                 train_config=training_config, dataset=dataset, datadir=args.datadir, local_seed=_seed,
                 global_seed=None, debug=args.debug, generate_sampling_profile=True, nsamples=nsamples[taskid],
                 portfolio_pth=None, cycle_portfolio=False, logger=_log,
                 opts={**fidelity.to_dict(), **opts})

        if i % (5 * max(ntasks // 100, 1)) == 0:
            # Report progress every time approximately 5% of the tasks have been fully pre-sampled and checkpoint
            # progress.
            _log.info(f"Worker {workerid}: Pre-sampling progress:\t{100 * i / ntasks:=6.2f}%\t({i}/{ntasks} tasks)")
            _log.info(f"Worker {workerid}: Maximum time per sampling operation: {max_time} seconds. Remaining Budget: "
                      f"{budget} seconds.")
            done.iloc[:i] = True
            shutil.copyfile(chkpt_filepath, f"{chkpt_filepath}.bak")
            done.to_pickle(chkpt_filepath)

        ## Update duration estimate
        end = time.time()
        duration = end - start
        max_time = max(max_time, duration)
        budget -= duration
        start = end

    _log.info(f"Worker {workerid}: Finished sub-program: initialize directory tree.")


# noinspection PyShadowingNames
def generate_jobs(args):
    _handle_debug(args)

    profile_path: Path = args.profile
    try:
        profile: pd.DataFrame = pd.read_pickle(profile_path)
        assert profile.index.is_unique, "Each row index in the profile must be unique throughout the profile."
        cpus_per_worker: pd.Series = profile.required.cpus
    except Exception as e:
        raise RuntimeError(f"Failed to read the sampling profile from {profile_path}.") from e

    if cpus_per_worker.nunique() != 1:
        _log.warning(f"The given profile is heterogeneous in the number of CPUs required per worker. The current state "
                     f"of this script does not support such profiles. It is advisable to break up the profile into "
                     f"multiple parts that are part-wise homogenous in the number of CPUs required per worker. ")

    cpus_per_worker: int = cpus_per_worker.max()
    assert isinstance(cpus_per_worker, int), f"The number of CPUs per worker must be an integer value, was " \
                                             f"{cpus_per_worker} of type {type(cpus_per_worker)}."

    job_config = sched_utils.JobConfig(
        cpus_per_worker=cpus_per_worker, cpus_per_node=args.cpus_per_node, nodes_per_job=args.nodes_per_job,
        timelimit=args.timelimit * 60
    )
    jobs = sched_utils.allocate_work(
        job_config=job_config, profile=profile, epochs=args.epochs,
        cpuh_utilization_cutoff=args.cpuh_utilization_cutoff, dynamic_timelimit=args.dynamic_timelimit,
        dynamic_nnodes=args.dynamic_nnodes, remainder_pth=args.profile_remainder, worker_id_offset=args.workerid_start,
        squish_configs=args.squish_configs
    )

    with open(args.template_dir / "job.template") as fp:
        job_template = Template(fp.read())

    with open(args.template_dir / "config.template") as fp:
        config_template = Template(fp.read())

    for jobid, job in enumerate(jobs, start=args.jobid_start):
        _log.info(f"Saving files for job #{jobid}.")
        job_name = f"job-{jobid}"
        rootdir = str(args.rootdir)

        _log.info(f"Saving portfolios.")
        portoflio_fn = job.save_worker_portfolios(jobid=jobid, portfolio_dir=args.portfolio_dir)

        job_str = job_template.substitute(
            rootdir=rootdir, script_dir=args.script_dir, job_name=job_name, **job.config.template_kwargs
        )
        srun_str = config_template.substitute(
            rootdir=rootdir, workerid_offset=job.worker_id_offset, portfolio=str(portoflio_fn),
            datadir=str(args.datadir), training_config_args=" ".join(_isolate_training_args(args))
        )

        _log.info("Saving job and config.")
        with open(args.script_dir / f"{job_name}.job", "w") as fp:
            fp.write(job_str)

        with open(args.script_dir / f"{job_name}.config", "w") as fp:
            fp.write(srun_str)

    _log.info(f"Finished generating all job files. To continue adding jobs to this job allocation, use the Job ID "
              f"Offset {args.jobid_start + len(jobs)}")


def argument_parser():
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument("--rootdir", type=Path,
                               help="This is the directory where the outputs of the jobs will be stored i.e. WORK. "
                                    "The base directory for each DirectoryTree will be "
                                    "'<rootdir>/<fidelity_dir>/tasks', where 'fidelity_dir' is generated by joining "
                                    "the names and values of the relevant fidelity parameters in a string separated by "
                                    "'-', e.g. 'N-1-W-8-Resolution-1.0'.")
    parent_parser.add_argument("--profile", type=Path,
                               help="Path to a *.pkl.gz file which will be either read from or written to, referring "
                                    "to the relevant sampling profile. When used with the command 'estimate', this "
                                    "file will be written to. When used with either 'init' or 'gen', it will be read "
                                    "from.")
    # TODO: This can be further optimized - it is only needed in the case when a model is actually trained
    parent_parser.add_argument("--datadir", type=Path,
                               help="The directory where all datasets are expected to be stored.")

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        title="Sub-commands", required=True,
        description="Pre-scheduling consists of three broad steps: 1. estimate- estimate the resource requirements and "
                    "sampling procedure, 2. initialize: initialize the directory tree for the estimated sampling "
                    "procedure, 3. generate: new bundled jobs to handle the actual evaluation. Steps 1 and 3 may be "
                    "repeated as more data becomes available."
    )

    ## Step 1 - Estimate Sampling Profile

    subparser_est = subparsers.add_parser(
        "estimate", aliases=["est"], parents=[parent_parser],
        help="Use the available metric data and desired computation parameters to estimate how many resources will be "
             "required for the desired evaluations as well as how the sampling procedure should be distributed to "
             "optimize resource utilization."
    )
    subparser_est.set_defaults(func=estimate_requirements)

    ## Step 2 - Initialize Directory Tree

    subparser_init = subparsers.add_parser(
        "initialize", aliases=["init"], parents=[parent_parser],
        help="Use the predicted total evaluation profile to generate the directory structure that will be needed for "
             "carrying out the expected calculations. Performing this step in advance allows the main jobs to run with "
             "the assumption that each individual model's parent directories already exist, fixed model "
             "configurations, training configurations and random seeds have been generated, and prevents a number of "
             "issues with multi-processing."
    )
    subparser_init.set_defaults(func=init_directory_tree)
    subparser_init.add_argument("--debug", action="store_true", help="Enable debug mode (very verbose) logging.")
    subparser_init.add_argument("--dataset",
                                choices=list(constants.Datasets.__members__.keys()),
                                help=f"The name of which dataset is to be used for model training and evaluation. Only "
                                     f"one of the provided choices can be used.")
    subparser_init.add_argument("--workerid", type=int,
                                help="An offset from 0 for this worker within the current job's allocation of workers. "
                                     "This does NOT correspond to a single 'taskid' but rather the portfolio of "
                                     "configurations that this worker will handle.")
    subparser_init.add_argument("--workerid-offset", type=int, default=0,
                                help="An additional fixed offset from 0 for this worker that, combined with the value "
                                     "of '--worker', gives the overall worker ID in the context of a job that has been "
                                     "split into multiple smaller parts.")
    subparser_init.add_argument("--nworkers", type=int,
                                help="The total number of workers that are expected to concurrently perform data "
                                     "verification in this root directory. This is used to coordinate the workers such "
                                     "that there is no overlap in their portfolios.")
    subparser_init.add_argument("--budget", type=int,
                                help="Time budget for each worker in seconds. This is used to approximate whether or "
                                     "not a worker has enough time to perform one more operation without any risk to "
                                     "any data. The script reserves 60 seconds of the budget for initialization, i.e. "
                                     "it always assumes a budget of 60 seconds less than what was specified here.")
    subparser_init.add_argument("--create_tree", action="store_true",
                                help="When given, also creates the required directory trees under rootdir. Otherwise, "
                                     "assumes that the directory tree already exists. This should be done exactly once "
                                     "with a single worker.")
    subparser_init.add_argument("opts", nargs=argparse.REMAINDER, default=None,
                                help="A variable number of optional keyword arguments provided as 2-tuples, each "
                                     "potentially corresponding to a hyper-parameter in the search space. If a match "
                                     "is found, that hyper-parameter is excluded from the search space and fixed to "
                                     "the given value instead. This also overrides the respective values read in from "
                                     "profiles.")

    ## Step 3 - Generate bundled jobs

    subparser_gen = subparsers.add_parser(
        "generate", aliases=["gen"], parents=[parent_parser],
        help="Use the existing resource requirement estimates and the given job resource availability parameters to "
             "generate concrete jobs that can be submitted to a SLURM cluster. This re-packages the estimated workload "
             "into job bundles of a desired size, e.g. 'n' nodes per job."
    )
    subparser_gen.set_defaults(func=generate_jobs)

    subparser_gen.add_argument("--cpus_per_node", type=int,
                               help="Job configuration - the number of CPUs that are to be expected in each node of "
                                    "the cluster.")
    subparser_gen.add_argument("--nodes_per_job", type=int,
                               help="The number of nodes that are available per independent job.")
    subparser_gen.add_argument("--timelimit", type=int,
                               help="The maximum amount of time (in minutes) that a single job is allowed to run for.")
    subparser_gen.add_argument("--dynamic_timelimit", action="store_true",
                               help="When this flag is given, the time limit of the job is dynamically adjusted to "
                                    "maximize the CPUh utilization. The dynamically adjusted time limit can only be "
                                    "lower than that specified by '--timelimit'. If this flag is omitted, the job will "
                                    "have exactly the value of '--timelimit' as its time limit.")
    subparser_gen.add_argument("--dynamic_nnodes", action="store_true",
                               help="When this flag is given, the number of nodes requested by each job is dynamically "
                                    "adjusted to maximize the CPUh utilization. The dynamically adjusted number of "
                                    "nodes can only be lower than that specified by '--nodes_per_job'. If this flag is "
                                    "omitted, each job will request exactly '--nodes_per_job' nodes, even if there "
                                    "aren't enough busy workers to use them.")
    subparser_gen.add_argument("--cpuh_utilization_cutoff", type=float, default=0.8,
                               help="Recommended minimum fraction of total allocated CPUh that should be actively used "
                                    "for computation. Generates a warning when the job allocation's expected CPUh "
                                    "utilization is below this value. Default: 0.8")
    subparser_gen.add_argument("--debug", action="store_true", help="Enable debug mode (very verbose) logging.")
    subparser_gen.add_argument("--template_dir", type=Path,
                               help="Path to a directory which contains templates using which the job files and their "
                                    "relevant srun config files are generated. The templates for jobs and configs "
                                    "should be named 'job.template' and 'config.template' respectively.")
    subparser_gen.add_argument("--script_dir", type=Path,
                               help="This is the directory where the generated job scripts and their respective srun "
                                    "configuration files will be stored.")
    subparser_gen.add_argument("--portfolio_dir", type=Path,
                               help="Path to a directory from where each worker will be able to store its own "
                                    "allocated portfolio of configurations to evaluate.")
    subparser_gen.add_argument("--profile_remainder", type=Path, default=None,
                               help="Full or relative (to the current working directory) path to *.pkl.gz file, in "
                                    "which any configurations from the profile that could not be fit into the given "
                                    "resource constraints will be stored. If the file exists, it is overwritten. If "
                                    "all configurations specified in the sampling profile have bee allocated to a "
                                    "worker's portfolio, this file is not created, but any existing file will remain"
                                    "untouched.")
    subparser_gen.add_argument("--jobid_start", type=int, default=0,
                                help="A fixed offset from 0 for all jobs produced by this run. This allows multiple "
                                     "executions of this script to be used to chain together multiple jobs without "
                                     "them interfering with each other by, e.g., continuously changing the input "
                                     "profile.")
    subparser_gen.add_argument("--workerid_start", type=int, default=0,
                                help="A fixed offset from 0 for the workers of all jobs produced by this run. This "
                                     "allows multiple executions of this script to be used to chain together multiple "
                                     "jobs without them interfering with each other by, e.g., continuously changing "
                                     "the input profile.")
    subparser_gen.add_argument("--squish_configs", action="store_true",
                               help="When this flag is given, any configs that do not fit into the allocated time "
                                    "budget will be squished into the job anyways so they can run for as long as "
                                    "possible.")

    # Unpack the training config into CLI flags.
    for k, v in constants.training_config.items():
        subparser_gen.add_argument(f"--{k}", **v)

    return parser


if __name__ == "__main__":
    ## Parse CLI
    args = argument_parser().parse_args()
    args.func(args)



import enum
import logging
import random
from itertools import repeat, cycle
from pathlib import Path
from typing import Dict, Union, Any, Optional, Iterable, Callable, List, Tuple, Hashable

import ConfigSpace
import jahs_bench.tabular.lib.naslib.utils.utils as naslib_utils
import numpy as np
import pandas as pd
import torch

from jahs_bench.tabular.lib.naslib.search_spaces.core.graph import Graph
from jahs_bench.tabular.lib.naslib.utils.utils import AttrDict, AverageMeter

import jahs_bench.tabular.lib.core.constants as constants
from jahs_bench.tabular.search_space import NASB201HPOSearchSpace
from jahs_bench.tabular.lib.core.custom_nasb201_code import CosineAnnealingLR


def get_training_config_help() -> dict:
    return {k: v["help"] for k, v in constants.training_config.items()}


def _query_config(config: Union[ConfigSpace.Configuration, Dict], param: str, default: Optional[Any] = None) -> Any:
    """ Query the given 'config' object for the parameter named 'param'. If the parameter is not found, returns default
    if given else None. This is necessary entirely because, as of the writing of this code, the 'default' argument of
     ConfigSpace.Configuration.get(), ConfigSpace v. 0.4.19, does not work as intended. """

    config = config.get_dictionary() if isinstance(config, ConfigSpace.Configuration) else config
    return config.get(param, default)


def _init_adam(model, config: Union[ConfigSpace.Configuration, Dict]):
    lr, weight_decay = _query_config(config, "LearningRate"), _query_config(config, "WeightDecay")
    optim = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)
    return optim


def _init_adamw(model, config: Union[ConfigSpace.Configuration, Dict]):
    lr, weight_decay = _query_config(config, "LearningRate"), _query_config(config, "WeightDecay")
    optim = torch.optim.AdamW(model.parameters(), lr=lr, weight_decay=weight_decay)
    return optim


def _init_sgd(model, config: Union[ConfigSpace.Configuration, Dict]):
    lr, momentum, weight_decay = _query_config(config, "LearningRate"), _query_config(config, "Momentum", 0.9), \
                                 _query_config(config, "WeightDecay")
    optim = torch.optim.SGD(model.parameters(), lr=lr, momentum=momentum, weight_decay=weight_decay, nesterov=True)
    return optim


@enum.unique
class Optimizers(enum.Enum):
    def __init__(self, constructor):
        self.constructor = constructor

    def construct(self, *args, **kwargs):
        return self.constructor(*args, **kwargs)

    SGD = _init_sgd,
    Adam = _init_adam,
    AdamW = _init_adamw,


class DirectoryTree(object):
    """ A very simply class that translates a base directory into various relevant output directories. Useful for
    maintaining consistency across the code.
    The following directory structure is used by DirectoryTree:

    <base_dir>
    |--> <task_id>/
    |    |--> metrics/
    |    |    |--> <elapsed_runtime>.pkl.gz    <-- these are metric DataFrame files
    |    |--> models/
    |    |    |--> <model_idx>/
    |    |    |    |--> metrics/
    |    |    |    |    |--> <elapsed_runtime>.pkl.gz    <-- these are metric DataFrame files
    |    |    |    |--> checkpoints/
    |    |    |    |    |--> <elapsed_runtime>.pt    <-- these are checkpoint files
    """

    def __init__(self, basedir: Path, taskid: Optional[int] = None, model_idx: Optional[int] = None,
                 read_only: bool = False):
        """
        The DirectoryTree enables a fixed, pre-determined directory sub-tree to be made and easily traversed at the
        given base directory, 'basedir'. 'taskid' and 'model_idx' are optional integers that can be changed and
        reassigned any time in order to generate the respective branch of the tree. Setting 'read_only' to True allows
        the tree to assume that no new directories need to be created if they don't exist and raise an Error if such an
        attempt to access such directory is made.
        """
        self.basedir = basedir.resolve()
        assert self.basedir.exists() and self.basedir.is_dir(), \
            f"The base directory and its parent directory tree must be created beforehand. Given base directory " \
            f"either does not exist or is not a directory: {str(self.basedir)}"
        self.read_only = read_only
        # Setting these to int values also creates the appropriate directories if read_only is False
        self.taskid = taskid
        self.model_idx = model_idx

    @property
    def taskid(self) -> int:
        return self._taskid

    @taskid.setter
    def taskid(self, new_id: Optional[int]):
        if new_id is not None:
            assert isinstance(new_id, int), f"Task ID must be an integer, was given {type(new_id)}"
            self._taskid = new_id
            if self.read_only:
                return
            if not self.task_dir.exists():
                self.task_dir.mkdir(parents=True)
            if not self.task_metrics_dir.exists():
                self.task_metrics_dir.mkdir(parents=True)
            if not self.task_models_dir.exists():
                self.task_models_dir.mkdir(parents=True)
        else:
            self._taskid = None

    @property
    def model_idx(self) -> int:
        return self._model_idx

    @model_idx.setter
    def model_idx(self, new_index: Optional[int]):
        if new_index is not None:
            assert isinstance(new_index, int), f"Model index must be an integer, was given {type(new_index)}"
            self._model_idx = new_index
            if self.read_only:
                return
            subdirs = [self.model_dir, self.model_metrics_dir, self.model_checkpoints_dir, self.model_tensorboard_dir]
            for subdir in subdirs:
                if not subdir.exists():
                    subdir.mkdir(parents=True)
        else:
            self._model_idx = None

    @property
    def task_dir(self) -> Path:
        return self.basedir / str(self.taskid)

    @property
    def task_config_file(self) -> Path:
        return self.task_dir / "task_config.json"

    @property
    def task_metrics_dir(self) -> Path:
        return self.task_dir / "metrics"

    @property
    def task_models_dir(self) -> Path:
        return self.task_dir / "models"

    @property
    def model_dir(self) -> Path:
        assert self.model_idx is not None, \
            "A valid model index needs to be set before the relevant model level subtree can be accessed."
        return self.task_models_dir / str(self.model_idx)

    @property
    def model_metrics_dir(self) -> Path:
        return self.model_dir / "metrics"

    @property
    def model_checkpoints_dir(self) -> Path:
        return self.model_dir / "checkpoints"

    @property
    def model_tensorboard_dir(self) -> Path:
        return self.model_dir / "tensorboard_logs"

    @property
    def model_error_description_file(self) -> Path:
        return self.model_dir / f"error_description.json"

    @property
    def existing_tasks(self) -> Optional[List[Path]]:
        return None if not self.basedir.exists() else [d for d in self.basedir.iterdir() if d.is_dir()]

    @property
    def existing_models(self) -> List[Path]:
        return None if not self.task_models_dir.exists() else [d for d in self.task_models_dir.iterdir() if d.is_dir()]


class SynchroTimer:
    """ A very simply class whose sole purpose is to provide a share-able object and interface so that multiple other
    pieces of code (such as Checkpointer and MetricLogger instances) can consult the same timing related logic. It
    handles the preparation of a signal based on the configuration of time/epoch frequency. Listeners that have been
    registered with a SynchroTimer object can then query it anytime to know if the signal has been prepared. This is
    by no means intended to be used for multi-process signal passing! """

    previous_timestamp: float
    last_epoch: int
    ping_interval_time: float
    ping_interval_epochs: float
    _signal: bool
    _listeners: Dict[Any, bool]

    def __init__(self, ping_interval_time: float = None, ping_interval_epochs: float = None):
        """ Setting either of the two interval types to None disables updates to the signal logic based on that
        property. It is thus possible to create a SynchroTimer object which can only generate signals at arbitrary
        intervals when manually asked to do so (see SynchroTimer.update()). """
        self.ping_interval_time = ping_interval_time
        self.ping_interval_epochs = ping_interval_epochs
        self.previous_timestamp = 0.
        self.last_epoch = -1
        self._listeners = {}
        self._signal = False

    def adjust(self, previous_timestamp: Optional[float] = None, last_epoch: Optional[int] = None):
        """ Sets the elapsed time and epochs without causing a signal to be generated. """
        self.previous_timestamp = previous_timestamp
        self.last_epoch = last_epoch

    def register_new_listener(self) -> Hashable:
        """ Allows SynchroTimer to know how many access points are listening to its signals and, in turn, ensure that
        each listener receives each unique instance of a signal only once. """
        id = len(self._listeners.keys()) + 1
        self._listeners[id] = self._signal
        return id

    def _update_signals_(self):
        """ Prepares signals for all _listeners. """
        for l in self._listeners.keys():
            self._listeners[l] = self._signal

    def update(self, timestamp: Optional[float] = None, curr_epoch: Optional[int] = None, force: Optional[bool] = False):
        """ Given an input of time/epochs or an overriding flag 'force', determines whether a signal should be
        prepared for the listeners or not. Even if ping_interval_time or ping_interval_epoch was set to None,
        the time and epochs can still be specified here to keep track of them. In such a case, however, they will not
        cause the timer's signal to be set. """

        time_signal = False if self.ping_interval_time is None \
            else (timestamp - self.previous_timestamp) >= self.ping_interval_time

        epoch_signal = False if self.ping_interval_epochs is None \
            else (curr_epoch - self.last_epoch) >= self.ping_interval_epochs

        self._signal = time_signal or epoch_signal or force
        if self._signal:
            self._update_signals_()
            self.previous_timestamp = timestamp
            self.last_epoch = curr_epoch

    def ping(self, id: Hashable):
        """ Any listener can call this function along with its registered ID to query the signal. """

        if id not in self._listeners:
            raise KeyError(f"Unrecognized listener ID {id}.")
        signal = self._listeners[id]

        # Either the signal was already False and remains unchanged, or was True and has now been consumed.
        self._listeners[id] = False
        return signal

class Checkpointer(object):
    """
    Essentially a stateful-function factory for checkpointing model training at discrete intervals of time and epochs.
    Initialized with references to all necessary objects and data for checkpointing and called by specifying the
    elapsed runtime and number of epochs. However, it can also be used to load existing checkpoints. Consult
    DirectoryTree for a breakdown of how the file structure is organized. A SynchroTimer object needs to be passed to
    the initializer argument 'timer' in order to enable logging functionality. Without it, a Checkpointer object can
    only be used in 'read-only' mode.
    """

    def __init__(self, model: Graph, optimizer: torch.optim.Optimizer, scheduler: CosineAnnealingLR,
                 dir_tree: DirectoryTree, timer: Optional[SynchroTimer] = None, logger: logging.Logger = None,
                 map_location=None):
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.dir_tree = dir_tree
        self.timer = timer
        self.runtime = 0. if self.timer is None else self.timer.previous_timestamp
        self.last_epoch = -1 if self.timer is None else self.timer.last_epoch
        self.logger = logger if logger is not None else logging.getLogger(__name__)
        logger.debug(f"Successfully initialized checkpointer.")
        self._load_latest(map_location)

    def __call__(self, force_checkpoint: bool = False):
        if self.timer is None:
            raise RuntimeError("The log functionality of Checkpointer is enabled only when a valid timer has been set.")
        if self._signal or force_checkpoint:
            runtime = self.timer.previous_timestamp
            last_epoch = self.timer.last_epoch
            output_file = self.dir_tree.model_checkpoints_dir / f"{runtime}.pt"
            torch.save(
                {
                    "model_state_dict": self.model.state_dict(),
                    "optimizer_state_dict": self.optimizer.state_dict(),
                    "scheduler_state_dict": self.scheduler.state_dict(),
                    "epochs": last_epoch,
                    "torch_rng_state": torch.get_rng_state(),
                    "numpy_rng_state": np.random.get_state(),
                    "python_rng_state": random.getstate(),
                },
                output_file
            )
            self.runtime = runtime
            self.last_epoch = last_epoch
            self.logger.debug(f"Checkpointed model training at f{str(output_file)}")

    @property
    def timer(self) -> SynchroTimer:
        return self._timer

    @timer.setter
    def timer(self, new_timer: SynchroTimer):
        if new_timer is None:
            self._timer = None
        elif not isinstance(new_timer, SynchroTimer):
            raise ValueError(f"Checkpointer requires a valid instance of SynchroTimer, cannot set timer to "
                             f"{type(new_timer)}.")
        else:
            self._timer = new_timer
            self._timer_id = self._timer.register_new_listener()

    @property
    def _signal(self):
        return False if self.timer is None else self.timer.ping(self._timer_id)

    @classmethod
    def _extract_runtime_from_filename(cls, f: Path) -> float:
        return float(f.name.rstrip(".pt"))

    @classmethod
    def _get_sorted_chkpt_paths(cls, pth: Path, ascending: bool = True) -> List[Path]:
        latest = sorted(
            pth.rglob("*.pt"),
            key=lambda f: cls._extract_runtime_from_filename(f),
            reverse=not ascending
        )
        return latest

    @classmethod
    def _load_checkpoint(cls, pths: Union[Path, List[Path]], safe_load: Optional[bool] = True,
                         map_location: Optional[Any] = None, get_path: bool = False) -> \
            Union[Optional[dict], Tuple[Optional[dict], Path]]:
        """ Attempts to load a checkpoint from either a single Path or a list of Paths, attempting each one in the
        order they occur. If 'safe_load' is set to False, the first checkpoint in the list that is readable from disk
        is loaded. It is set to True by default - i.e. if the first checkpoint cannot be loaded, an error is raised.
        'map_location' has been provided for compatibility with GPUs. """

        if isinstance(pths, Path):
            pths = [pths]

        state_dicts = None
        pth = None
        for pth in pths:
            try:
                state_dicts = torch.load(pth, map_location=map_location)
            except Exception as e:
                if safe_load:
                    raise RuntimeError(f"Could not safely load checkpoint at {pth}") from e
                self.logger.info(f"Found possibly corrupt checkpoint at {pth}, reverting to an earlier checkpoint.")
            else:
                break

        return (state_dicts, pth) if get_path else state_dicts

    def _load_latest(self, safe_load: Optional[bool] = True, map_location: Optional[Any] = None):
        """ Attempts to load a previously saved checkpoint in order to resume model training. If a checkpoint is
        successfully loaded, the model, optimizer and scheduler state dictionaries are appropriately loaded and the
        relevant values of runtime and last_epoch are updated, else, the state dictionaries and other object
        attributes are left untouched.  If 'safe_load' is set to False, the most recent checkpoint that is readable
        from disk is loaded. Be careful with this setting - it is possible for the loaded checkpoint data and metrics
        data to be out of sync in such a scenario. It is set to True by default - i.e. if the latest checkpoint cannot
        be loaded, a RuntimeError is raised. 'map_location' has been provided for compatibility with GPUs. """

        latest = self._get_sorted_chkpt_paths(self.dir_tree.model_checkpoints_dir, ascending=False)
        state_dicts, pth = self._load_checkpoint(pths=latest, safe_load=safe_load, map_location=map_location,
                                                 get_path=True)

        if state_dicts is None:
            self.logger.info(f"No valid checkpoints found at {self.dir_tree.model_checkpoints_dir}.")
            return

        # state_dicts = torch.load(latest, map_location=map_location)
        self.model.load_state_dict(state_dicts["model_state_dict"])
        self.optimizer.load_state_dict(state_dicts["optimizer_state_dict"])
        self.scheduler.load_state_dict(state_dicts["scheduler_state_dict"])
        torch_rng_state = state_dicts["torch_rng_state"]
        # Fix needed for compatibilty with cuda, not yet tested for repeatability!!
        if "cpu" not in str(torch_rng_state.device.type):
            torch_rng_state = torch_rng_state.cpu()
        torch.set_rng_state(torch_rng_state)
        np.random.set_state(state_dicts["numpy_rng_state"])
        random.setstate(state_dicts["python_rng_state"])
        self.last_epoch = state_dicts["epochs"]
        self.runtime = self._extract_runtime_from_filename(pth)
        self.logger.info(f"Loaded existing checkpoint from {str(pth)}")


class MetricLogger(object):
    """
    Holds a set of metrics, information about where they are to be logged, the frequency at which they should be
    logged as well as some functionality to convert any given set of metrics into a pandas DataFrame. Consult
    DirectoryTree for a breakdown of how the file structure is organized. A SynchroTimer object needs to be passed to
    the initializer argument 'timer' in order to enable logging functionality. Without it, a MetricLogger object can
    only be used in 'read-only' mode.
    """

    @enum.unique
    class MetricSet(enum.Enum):
        """ Defines uniquely recognized sets of metrics that are handled slightly differently from the norm. """
        model = enum.auto()
        task = enum.auto()
        default = enum.auto()

    def __init__(self, dir_tree: DirectoryTree, metrics: Dict, timer: Optional[SynchroTimer] = None,
                 set_type: MetricSet = MetricSet.default, logger: logging.Logger = None, where: Path = None,
                 safe_load: bool = True):
        """
        Setup a MetricLogger instance.

        :param dir_tree: DirectoryTree
            An instance of DirectoryTree that has already been initialized with the parameters appropriate for usage
            with this object, i.e. base directory, taskid and model index depending on which level of the tree this
            MetricLogger is intended to interface with.
        :param metrics: dict-like
            A dict-like object that will store the actual metric data. This object can be initialized with keys and no
            data before being passed to the MetricLogger, in which case the data will be filled in by loading it from
            disk.
        :param timer: SynchroTimer
            Used to coordinate the timing of saving metrics to disk with other parts of the code.
        :param set_type: Enum
            One of MetricLogger.MetricSet to indicate which level of the directory this logger will be interfacing with.
        :param logger: logging.Logger
            Used to generate log messages. Optional.
        :param where: Path
            An optional Path to a directory containing *.pkl.gz files, which should be compatible pickled Pandas
            DataFrames used to initialize the metrics dict. If None, the directory is read from the 'dir_tree'.
        :param safe_load: bool
            If True (default), failing to load the latest metrics log raises a RuntimeError. Otherwise, the next most
            recent readable metrics log is used instead.
        """
        self.dir_tree = dir_tree
        self.metrics = metrics
        self.timer = timer
        self.set_type = set_type
        self.logger = logger if logger is not None else logging.getLogger(__name__)
        self.elapsed_runtime = 0. if self.timer is None else self.timer.previous_timestamp
        self.resume_latest_saved_metrics(where=where, safe_load=safe_load)

    @property
    def timer(self) -> SynchroTimer:
        return self._timer

    @timer.setter
    def timer(self, new_timer: SynchroTimer):
        if new_timer is None:
            self._timer = None
        elif not isinstance(new_timer, SynchroTimer):
            raise ValueError(f"Checkpointer requires a valid instance of SynchroTimer, cannot set timer to "
                             f"{type(new_timer)}.")
        else:
            self._timer = new_timer
            self._timer_id = self._timer.register_new_listener()

    @property
    def _signal(self):
        return False if self.timer is None else self.timer.ping(self._timer_id)

    @classmethod
    def _nested_dict_to_df(cls, nested: Union[dict, Any]) -> pd.DataFrame:
        """ Given a nested dict of dicts, use each level of the nested keys as column index levels in a pandas
        MultiIndex and convert the entire data structure into a pandas DataFrame. Note that any item that is not a dict
        will be considered to be a leaf node in the nested dict structure, therefore, any iterables will be directly
        converted into columns. Care should be taken to make sure that such iterables are of an appropriate length.
        Single data items at each level will only be copied over to as many indices as the length of the longest
        iterable at that level. """

        assert isinstance(nested, dict), f"Expected input to be a possibly nested dictionary, received {type(nested)}."

        index = None
        new_df = {}
        singles = {}
        for k, v in nested.items():
            if isinstance(v, dict):
                new_df[k] = cls._nested_dict_to_df(v)
            elif isinstance(v, Iterable):
                new_df[k] = pd.Series(v)
            else:
                singles[k] = v
                continue
            if index is None:
                index = new_df[k].index
                # new_df = pd.DataFrame(index=vdf.index)
        if new_df:
            new_df = pd.concat(new_df, axis=1)  # Create new level in columns MultiIndex or promote Index to MultiIndex
            new_df = new_df.assign(**singles)
        else:
            new_df = pd.DataFrame(singles, index=[0])

        return new_df

    @classmethod
    def _df_to_nested_dict(cls, df: pd.DataFrame) -> dict:
        """ Given a DataFrame, attempts to convert it into a nested dict. This is intended to be the inverse of
        _nested_dict_to_df() and therefore only intended to work with DataFrame objects that were originally created by
        that function. """

        metrics_dict = AttrDict()
        for k in df.columns.unique(0):
            if isinstance(df[k], pd.Series):
                metrics_dict[k] = df[k].to_list()
            else:
                metrics_dict[k] = cls._df_to_nested_dict(df[k])

        return metrics_dict

    @classmethod
    def _handle_model_metric_dict(cls, metrics: Dict) -> pd.DataFrame:
        df = cls._nested_dict_to_df(metrics)
        assert isinstance(df, pd.DataFrame), f"This function should only be used with a DataFrame, not {type(df)}"
        assert isinstance(df.index, pd.Index), \
            f"Expected the model-level metrics DataFrame index to be an Index, not {type(df.index)}"
        assert isinstance(df.columns, pd.MultiIndex), \
            f"Expected the model-level metrics DataFrame columns to be a MultiIndex, not {type(df.columns)}"
        assert df.columns.nlevels == len(constants.metricdf_column_level_names), \
            f"Expected the model-level metrics DataFrame to have exactly {len(constants.metricdf_column_level_names)} levels " \
            f"in the columns MultiIndex, got {df.columns.nlevels}. Consult tabular_sampling.lib.constants for more " \
            f"details about the expected structure."

        df.index = df.index.set_names([constants.MetricDFIndexLevels.epoch.value]) + 1
        df.columns.set_names(constants.metricdf_column_level_names, inplace=True)
        return df

    @classmethod
    def _handle_task_metric_dict(cls, metrics: Dict) -> pd.DataFrame:
        model_configs = pd.DataFrame(metrics["model_config"]).to_dict()
        model_configs = {k: list(v.values()) for k, v in
                         model_configs.items()}  # Convert list of dicts to dict of lists
        df = cls._nested_dict_to_df({**metrics, **{"model_config": model_configs}})

        assert isinstance(df, pd.DataFrame), f"This function should only be used with a DataFrame, not {type(df)}"
        assert isinstance(df.index, pd.Index), \
            f"Expected the task-level metrics DataFrame index to be an Index, was {type(df.index)}"
        assert isinstance(df.columns, pd.MultiIndex), \
            f"Expected the task-level metrics DataFrame columns to be a MultiIndex, not {type(df.columns)}"
        assert df.columns.nlevels == 2, \
            f"Expected the task-level metrics DataFrame to have exactly 2 levels in the columns MultiIndex, got " \
            f"{df.columns.nlevels}"

        return df

    @property
    def _metric_set_dict_handlers(self):
        return {
            MetricLogger.MetricSet.model: self._handle_model_metric_dict,
            MetricLogger.MetricSet.task: self._handle_task_metric_dict,
            MetricLogger.MetricSet.default: self._nested_dict_to_df
        }

    @classmethod
    def _handle_task_metric_df(cls, df: pd.DataFrame) -> dict:
        # Handle model configs specially since it's the only column index that uses a MultiIndex and needs to be
        # converted to a list of dicts format.
        model_configs = list(df["model_config"].transpose().to_dict().values())
        # The following is necessary to undo the automatic conversion of nested dicts to MultiIndex columns.
        other_cols = df.columns.unique(0).difference(["model_config"])
        metric_dict = cls._df_to_nested_dict(df[other_cols].droplevel(1, axis=1))
        metric_dict["model_config"] = model_configs

        return metric_dict

    @property
    def _metric_set_df_handlers(self):
        return {
            MetricLogger.MetricSet.model: self._df_to_nested_dict,
            MetricLogger.MetricSet.task: self._handle_task_metric_df,
            MetricLogger.MetricSet.default: self._df_to_nested_dict
        }

    def _generate_df(self) -> pd.DataFrame:
        """ Converts the currently stored dict of metrics into a Pandas DataFrame ready for being saved to disk. """

        # metric_df = self._nested_dict_to_df(self.metrics)
        # noinspection PyArgumentList
        metric_df = self._metric_set_dict_handlers[self.set_type](self.metrics)
        return metric_df

    @property
    def _metric_set_to_log_directory(self):
        return {
            MetricLogger.MetricSet.model: "model_metrics_dir",
            MetricLogger.MetricSet.task: "task_metrics_dir",
            MetricLogger.MetricSet.default: "basedir"
        }

    @property
    def log_directory(self):
        return getattr(self.dir_tree, self._metric_set_to_log_directory[self.set_type])

    def log(self, force: bool = False, where: Optional[Path] = None):
        if self.timer is None:
            raise RuntimeError("The log functionality of MetricLogger is enabled only when a valid timer has been set.")
        if self._signal or force:
            elapsed_runtime = self.timer.previous_timestamp
            df = self._generate_df()
            pth = where if where is not None else self.log_directory
            df.to_pickle(pth / f"{elapsed_runtime}.pkl.gz")
            self.elapsed_runtime = elapsed_runtime
            self.logger.info(f"Logged metrics at {str(pth)}")

    @classmethod
    def _extract_runtime_from_filename(cls, f: Path) -> float:
        return float(f.name.rstrip(".pkl.gz"))

    @classmethod
    def _get_sorted_metric_paths(cls, pth: Path, ascending: bool = True) -> List[Path]:
        latest = sorted(
            pth.rglob("*.pkl.gz"),
            key=lambda f: cls._extract_runtime_from_filename(f),
            reverse=not ascending
        )
        return latest

    @classmethod
    def _load_metrics_log(cls, pths: Union[Path, List[Path]],
                          safe_load: Optional[bool] = True, get_path: bool = False) -> \
            Union[Optional[pd.DataFrame], Tuple[Optional[pd.DataFrame], Path]]:
        """ Attempts to load a metrics log from either a single Path or a list of Paths, attempting each one in the
        order they occur. If 'safe_load' is set to False, the first metrics log in the list that is readable from disk
        is loaded. It is set to True by default - i.e. if the first metrics log cannot be loaded, a RuntimeError is
        raised. """

        if isinstance(pths, Path):
            pths = [pths]

        metric_df = None
        pth = None
        for pth in pths:
            try:
                metric_df = pd.read_pickle(pth)
            except Exception as e:
                if safe_load:
                    raise RuntimeError(f"Could not safely load metric DataFrame at {pth}") from e

                self.logger.info(f"Found possibly corrupt metric DataFrame at {pth}, reverting to an earlier "
                                 f"checkpoint.")
            else:
                break

        return (metric_df, pth) if get_path else metric_df

    def resume_latest_saved_metrics(self, where: Optional[Path] = None, safe_load: Optional[bool] = True):
        """ Used in conjunction with the corresponding method of Checkpointer to resume model training and evaluation
        from where it was left off. If 'safe_load' is set to False, the most recent metric data that is readable from
        disk is loaded. Be careful with this setting - it is possible for the loaded checkpoint data and metrics data
        to be out of sync in such a scenario. It is set to True by default - i.e. if the latest metrics DataFrame
        cannot be loaded, an error is raised."""

        pth = where if where is not None else self.log_directory
        latest = self._get_sorted_metric_paths(pth=pth, ascending=False)
        metric_df, dfpth = self._load_metrics_log(pths=latest, safe_load=safe_load, get_path=True)

        if metric_df is None:
            self.logger.info(f"No valid pre-existing metric DataFrames found at {str(pth)}.")
            return

        # noinspection PyArgumentList
        loaded_metrics = self._metric_set_df_handlers[self.set_type](metric_df)
        for k in self.metrics.keys():
            self.metrics[k] = loaded_metrics[k]

        self.elapsed_runtime = self._extract_runtime_from_filename(dfpth)


def attrdict_factory(metrics: Optional[List[str]] = None, template: Callable = AverageMeter) -> AttrDict:
    """ Convenience function for generating an AttrDict object with arbitrary keys initialized using a factory
    function, essentially an AttrDict extension to collections.defaultdict. 'metrics' is a list of keys (usually
    strings) and the initial value of these keys can optionally be set by passing an object factory to the parameter
    'template', which defaults to naslib.utils.utils.AverageMeter. """

    metrics = AttrDict({m: template() for m in metrics} if metrics else {})
    return metrics


def default_global_seed_gen(rng: Optional[np.random.RandomState] = None, global_seed: Optional[int] = None) \
        -> Iterable[int]:
    if global_seed is not None:
        return global_seed if isinstance(global_seed, Iterable) else repeat(global_seed)
    elif rng is not None:
        def seeds():
            while True:
                yield rng.randint(0, 2 ** 32 - 1)

        return seeds()
    else:
        raise ValueError("Cannot generate sequence of global seeds when both 'rng' and 'global_seed' are None.")


# TODO: Update documentation
def adapt_search_space(original_space: Union[NASB201HPOSearchSpace, ConfigSpace.ConfigurationSpace],
                       portfolio: Optional[pd.DataFrame] = None, taskid: Optional[int] = None,
                       opts: Optional[Union[Dict[str, Any], List[str]]] = None, suffix: Optional[str] = "_custom") -> \
        Union[Tuple[ConfigSpace.ConfigurationSpace, bool], bool]:
    """ Given a NASB201HPOSearchSpace object and a valid configuration, restricts the respective configuration space
    object and consequently the overall search space by setting the corresponding parameters to constant values. Such
    a configuration may be provided either by means of a portfolio file along with the relevant taskid or a dictionary
    'opts' mapping strings to values of any type, or both in which case values in 'opts' take precedence. In both
    cases, the configuration is read as a dictionary and the function attempts to autonomously restrict the search
    space such that for each key *k*, the corresponding parameter is set to the constant value *v*. An optional
    string 'suffix' may be provided to modify the name of the ConfigSpace object. Default: '_custom'. If None, the
    name remains unchanged. The name also remains unchanged in the case when the given space was not modified at all,
    most likely because no valid keys were present in the configuration dictionary or the dictionary was empty. In
    such a case, False is returned. Otherwise, True is returned. """

    new_consts = {}
    if portfolio is None and opts is None:
        return False
    elif portfolio is not None:
        if taskid is None:
            raise ValueError(f"When a portfolio is given, an integer taskid must also be provided. Was given {taskid}")
        else:
            new_consts = portfolio.loc[taskid % portfolio.index.size, :].to_dict()

    # Convert opts from list of string pairs to a mapping from string to string.
    if isinstance(opts, List):
        i = iter(opts)
        opts = {k: v for k, v in zip(i, i)}
    new_consts = {**new_consts, **opts}

    if hasattr(original_space, "config_space"):
        config_space = original_space.config_space
        flag_cs_attr = True
    else:
        config_space = original_space
        flag_cs_attr = False

    known_params = {p.name: p for p in config_space.get_hyperparameters()}

    def param_interpretor(param, value):
        known_config_space_value_types = {
            ConfigSpace.UniformIntegerHyperparameter: int,
            ConfigSpace.UniformFloatHyperparameter: float,
            ConfigSpace.CategoricalHyperparameter: lambda x: type(param.choices[0])(x),
            ConfigSpace.OrdinalHyperparameter: lambda x: type(param.sequence[0])(x),
        }
        return known_config_space_value_types[type(param)](value)

    modified = False

    for arg, val in new_consts.items():
        if arg in known_params:
            old_param = known_params[arg]
            new_val = param_interpretor(old_param, val)
            if isinstance(new_val, bool):
                new_val = int(new_val)  # Because ConfigSpace doesn't allow Boolean constants.
            known_params[arg] = ConfigSpace.Constant(arg, new_val,
                                                     meta=dict(old_param.meta, **dict(constant_overwrite=True)))
            modified = True

    if modified:
        new_config_space = ConfigSpace.ConfigurationSpace(f"{config_space.name}{suffix if suffix is not None else ''}")
        new_config_space.add_hyperparameters(known_params.values())
        if flag_cs_attr:
            original_space.config_space = new_config_space
        else:
            return new_config_space, modified

    return modified


def default_random_sampler(search_space: NASB201HPOSearchSpace, global_seed_gen: Iterable[int],
                           rng: np.random.RandomState, **kwargs) -> Tuple[NASB201HPOSearchSpace, dict, int]:
    """ Given a compatible search space, a global seed generator for fixing global seeds before model training and a
    local RNG for controlling sampling from the search space, generates samples from the search space, their
    configuration dictionary and the corresponding global seed. In this process, the global seed is automatically set.
    The extra keyword arguments are complete ignored and have been provided only to allow function signature
    overloading.
    """
    while True:
        curr_global_seed = next(global_seed_gen)
        naslib_utils.set_seed(curr_global_seed)
        model: NASB201HPOSearchSpace = search_space.clone()
        model.sample_random_architecture(rng=rng)
        model_config = model.config.get_dictionary()
        yield model, model_config, curr_global_seed


def model_sampler(search_space: NASB201HPOSearchSpace, taskid: int, global_seed_gen: Iterable[int],
                  rng: np.random.RandomState, portfolio_pth: Optional[Path] = None, opts: Optional[dict] = None,
                  cycle_models: bool = False, **kwargs) -> Iterable[Tuple[NASB201HPOSearchSpace, dict, int]]:
    """ Generates an iterable over tuples of initialized models, their configuration dictionary, and the global int
    seed used to initialize the global RNG state of PyTorch, NumPy and Random for that model by parsing a given search
    space, taking into account an optional portfolio of configurations. 'global_seed_gen' is an iterable of integers
    used to generate the int seeds for each model and 'rng' is used for consistent random sampling. An optional
    dictionary 'opts' mapping parameter names to constant values can be used to override the corresponding parameter
    values in the portfolio and the search space. If the flag 'cycle_models' is set, then in the case when a portfolio
    of fixed model configs for each task is given, the configs are cycled infinitely, thereby allowing the same configs
    to be evaluated under multiple global seeds. This flag has no effect when the portfolio does not contains
    model-specific configs or when no portfolio is given. Warning: When using the fixed sampling mode by specifying
    each individual model config, the behaviour of specifying additional 'opts' is undefined. """

    taskid_lvl = constants.MetricDFIndexLevels.taskid.value
    use_fixed_sampler = False
    if opts is None:
        opts = {}
    elif isinstance(opts, list):
        i = iter(opts)
        opts = {k: v for k, v in zip(i, i)}

    if portfolio_pth is not None:
        # Accommodate for a portfolio file modifying the search space.
        portfolio = pd.read_pickle(portfolio_pth)
        if isinstance(portfolio.index, pd.MultiIndex):
            # Format 2 of the portfolio: We need to sample only the models specified in the portfolio
            portfolio = portfolio.xs(taskid % portfolio.index.unique(taskid_lvl).size, axis=0, level=taskid_lvl)
            if "opts" in portfolio.columns.levels[0]:
                # These values should be constant across all sampled models
                opts = {**portfolio["opts"].iloc[0].to_dict(), **opts}
            use_fixed_sampler = True
            portfolio = portfolio["model_config"]
        else:
            # Format 1 of the portfolio: We need to fix one or more parameters in the search space for this task
            opts = {**portfolio.loc[taskid % portfolio.index.unique(taskid_lvl).size, :].to_dict(), **opts}

    # Use the search space modifiers that have now already subsumed the portfolio
    adapt_search_space(search_space, opts=opts, suffix=f"_task_{taskid}")

    if use_fixed_sampler:
        # The portfolio now consists of only the model configs that we want to iterate over within this task.
        model_inds = cycle(portfolio.index.values) if cycle_models else portfolio.index.values

        def sampler():
            for idx, curr_global_seed in zip(model_inds, global_seed_gen):
                model_config = portfolio.loc[idx, :].to_dict()
                model = search_space.clone()
                model.clear()
                model.config = ConfigSpace.Configuration(
                    model.config_space,
                    {**model.config_space.get_default_configuration().get_dictionary(), **model_config}
                )
                model._construct_graph()
                model_config = model.config.get_dictionary()
                yield model, model_config, curr_global_seed

        return sampler()
    else:
        return default_random_sampler(search_space=search_space, global_seed_gen=global_seed_gen, rng=rng)





"""
Adapted in from the original NASBench-201 code repository at 
https://github.com/D-X-Y/AutoDL-Projects/tree/bc4c4692589e8ee7d6bab02603e69f8e5bd05edc
"""

import copy, torch
import torch.nn as nn
import numpy as np


def get_model_flops(model, input_shape, transfer_devices=False, device="cpu"):
    """ For a given model and input shape, estimates the number of floating point operations that will be performed
    per second of computation, counted in millions, i.e. a value of 1. indicates approximately 1 million floating
    point operations per second. Note that calls to this function do advance the state of the PyTorch PRNG. """
    model = add_flops_counting_methods(model)
    model.eval()

    cache_inputs = torch.rand(*input_shape)
    if transfer_devices: 
        cache_inputs = cache_inputs.to(device)
    with torch.no_grad():
        _____ = model(cache_inputs)
    FLOPs = compute_average_flops_cost(model) / 1e6

    torch.cuda.empty_cache()
    model.apply(_remove_hook_function)
    return FLOPs


# ---- Public functions
def add_flops_counting_methods(model):
    model.__batch_counter__ = 0
    _add_batch_counter_hook_function(model)
    model.apply(_add_flops_counter_variable_or_reset)
    model.apply(_add_flops_counter_hook_function)
    return model


def compute_average_flops_cost(model):
    """
    A method that will be available after add_flops_counting_methods() is called on a desired net object.
    Returns current mean flops consumption per image.
    """
    batches_count = model.__batch_counter__
    flops_sum = 0
    # or isinstance(module, torch.nn.AvgPool2d) or isinstance(module, torch.nn.MaxPool2d) \
    for module in model.modules():
        if isinstance(module, torch.nn.Conv2d) or isinstance(module, torch.nn.Linear) \
                or isinstance(module, torch.nn.Conv1d) \
                or hasattr(module, 'calculate_flop_self'):
            flops_sum += module.__flops__
    return flops_sum / batches_count


# ---- Internal functions
def _pool_flops_counter_hook(pool_module, inputs, output):
    batch_size = inputs[0].size(0)
    kernel_size = pool_module.kernel_size
    out_C, output_height, output_width = output.shape[1:]
    assert out_C == inputs[0].size(1), '{:} vs. {:}'.format(out_C, inputs[0].size())

    overall_flops = batch_size * out_C * output_height * output_width * kernel_size * kernel_size
    pool_module.__flops__ += overall_flops


def _self_calculate_flops_counter_hook(self_module, inputs, output):
    overall_flops = self_module.calculate_flop_self(inputs[0].shape, output.shape)
    self_module.__flops__ += overall_flops


def _fc_flops_counter_hook(fc_module, inputs, output):
    batch_size = inputs[0].size(0)
    xin, xout = fc_module.in_features, fc_module.out_features
    assert xin == inputs[0].size(1) and xout == output.size(1), 'IO=({:}, {:})'.format(xin, xout)
    overall_flops = batch_size * xin * xout
    if fc_module.bias is not None:
        overall_flops += batch_size * xout
    fc_module.__flops__ += overall_flops


def _conv1d_flops_counter_hook(conv_module, inputs, outputs):
    batch_size = inputs[0].size(0)
    outL = outputs.shape[-1]
    [kernel] = conv_module.kernel_size
    in_channels = conv_module.in_channels
    out_channels = conv_module.out_channels
    groups = conv_module.groups
    conv_per_position_flops = kernel * in_channels * out_channels / groups

    active_elements_count = batch_size * outL
    overall_flops = conv_per_position_flops * active_elements_count

    if conv_module.bias is not None:
        overall_flops += out_channels * active_elements_count
    conv_module.__flops__ += overall_flops


def _conv2d_flops_counter_hook(conv_module, inputs, output):
    batch_size = inputs[0].size(0)
    output_height, output_width = output.shape[2:]

    kernel_height, kernel_width = conv_module.kernel_size
    in_channels = conv_module.in_channels
    out_channels = conv_module.out_channels
    groups = conv_module.groups
    conv_per_position_flops = kernel_height * kernel_width * in_channels * out_channels / groups

    active_elements_count = batch_size * output_height * output_width
    overall_flops = conv_per_position_flops * active_elements_count

    if conv_module.bias is not None:
        overall_flops += out_channels * active_elements_count
    conv_module.__flops__ += overall_flops


def _batch_counter_hook(module, inputs, output):
    # Can have multiple inputs, getting the first one
    inputs = inputs[0]
    batch_size = inputs.shape[0]
    module.__batch_counter__ += batch_size


def _add_batch_counter_hook_function(module):
    if not hasattr(module, '__batch_counter_handle__'):
        handle = module.register_forward_hook(_batch_counter_hook)
        module.__batch_counter_handle__ = handle


def _add_flops_counter_variable_or_reset(module):
    if isinstance(module, torch.nn.Conv2d) or isinstance(module, torch.nn.Linear) \
            or isinstance(module, torch.nn.Conv1d) \
            or isinstance(module, torch.nn.AvgPool2d) or isinstance(module, torch.nn.MaxPool2d) \
            or hasattr(module, 'calculate_flop_self'):
        module.__flops__ = 0


def _add_flops_counter_hook_function(module):
    if isinstance(module, torch.nn.Conv2d):
        if not hasattr(module, '__flops_handle__'):
            handle = module.register_forward_hook(_conv2d_flops_counter_hook)
            module.__flops_handle__ = handle
    elif isinstance(module, torch.nn.Conv1d):
        if not hasattr(module, '__flops_handle__'):
            handle = module.register_forward_hook(_conv1d_flops_counter_hook)
            module.__flops_handle__ = handle
    elif isinstance(module, torch.nn.Linear):
        if not hasattr(module, '__flops_handle__'):
            handle = module.register_forward_hook(_fc_flops_counter_hook)
            module.__flops_handle__ = handle
    elif isinstance(module, torch.nn.AvgPool2d) or isinstance(module, torch.nn.MaxPool2d):
        if not hasattr(module, '__flops_handle__'):
            handle = module.register_forward_hook(_pool_flops_counter_hook)
            module.__flops_handle__ = handle
    elif hasattr(module, 'calculate_flop_self'):  # self-defined module
        if not hasattr(module, '__flops_handle__'):
            handle = module.register_forward_hook(_self_calculate_flops_counter_hook)
            module.__flops_handle__ = handle


def _remove_hook_function(module):
    hookers = ['__batch_counter_handle__', '__flops_handle__']
    for hooker in hookers:
        if hasattr(module, hooker):
            handle = getattr(module, hooker)
            handle.remove()
    keys = ['__flops__', '__batch_counter__', '__flops__'] + hookers
    for ckey in keys:
        if hasattr(module, ckey): delattr(module, ckey)



import requests
import tarfile
from pathlib import Path

surrogate_url = "https://ml.informatik.uni-freiburg.de/research-artifacts/jahs_bench_201/v1.0.0/assembled_surrogates.tar"
metric_url = "https://ml.informatik.uni-freiburg.de/research-artifacts/jahs_bench_201/v1.0.0/metric_data.tar"


def download_and_extract_url(url, save_dir, filename):
    save_dir = Path(save_dir)
    save_dir.mkdir(parents=True, exist_ok=True)
    save_tar_file = save_dir / filename

    print(f"Starting download of {url}, this might take a while.")
    with requests.get(url, stream=True) as response:
        with open(save_tar_file, 'wb') as f:
            f.write(response.raw.read())

    print("Download finished, extracting now")
    with tarfile.open(save_tar_file, 'r') as f:
        f.extractall(path=save_dir)
    print("Done extracting")


def download_surrogates(save_dir="jahs_bench_data"):
    download_and_extract_url(surrogate_url, save_dir, "assembled_surrogates.tar")


def download_metrics(save_dir="jahs_bench_data"):
    download_and_extract_url(metric_url, save_dir, "metric_data.tar")


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--target", default="surrogates", choices=["surrogates", "metrics", "all"])
    parser.add_argument("--save_dir", default="jahs_bench_data")
    args = parser.parse_args()

    if args.target == "surrogates":
        download_surrogates(args.save_dir)
    elif args.target == "metrics":
        download_metrics(args.save_dir)
    else:
        download_surrogates(args.save_dir)
        download_metrics(args.save_dir)



import argparse
import logging
from pathlib import Path
from typing import Optional, Sequence
import yaml

import pandas as pd
import scipy.stats
import sklearn.metrics
from jahs_bench.surrogate import model

_default_test_set_fn = "test_set.pkl.gz"
_default_test_pred_fn = "test_pred.pkl.gz"
_log = logging.getLogger(__name__)


def load_test_set(testset_file: Path, outputs: Optional[Sequence[str]] = None) -> \
                  pd.DataFrame:
    assert testset_file.exists()
    _log.info(f"Attempting to read test set from file {testset_file}.")
    test_set = pd.read_pickle(testset_file)
    if outputs is not None:
        _log.info(f"Restricting test set to the given outputs: {outputs}.")
        outputs = outputs if isinstance(outputs, list) else [outputs] \
            if isinstance(outputs, str) else list(outputs)  # Coerce into a list

        assert test_set["labels"].columns.intersection(outputs).size == len(outputs), \
            f"The given set of outputs is not a subset of the outputs present in the " \
            f"test set: {test_set['labels'].columns.tolist()}."

        sel = test_set[["features"]].columns.tolist() + \
              test_set.loc[:, ("labels", outputs)].columns.tolist()
        test_set = test_set.loc[:, sel]

    _log.info(f"Successfully loaded test set of shape {test_set.shape}.")
    return test_set


def evaluate_test_set(test_set: pd.DataFrame, model_dir: Path,
                      outputs: Optional[Sequence[str]] = None,
                      save_dir: Optional[Path] = None) -> pd.DataFrame:
    """ Evaluate the model on the given test set and returns the predictions as a pandas
    DataFrame. """

    _log.info("Extracting the input features and expected labels from the test set.")
    xtest = test_set.loc[:, "features"]
    ytest = test_set.loc[:, "labels"]
    assert xtest.shape[0] == ytest.shape[0]

    _log.info(f"Test set has {xtest.shape[0]} data points.")

    if outputs is None:
        outputs = ytest.columns.tolist()

    ypred = {}
    for output in outputs:
        _log.info(f"Preparing to load surrogate for output: {output}")
        model_subdir = model_dir / str(output)
        surrogate = model.XGBSurrogate.load(model_subdir)

        _log.info(f"Generating surrogate predictions.")
        ypred[output] = surrogate.predict(xtest)

    _log.info("Concatenating predictions.")
    ypred = pd.concat(ypred, axis=1)
    if save_dir is not None:
        # TODO: If 'exist_ok=False' is to be used, it should be done at the very start
        save_dir.mkdir(exist_ok=True, parents=True)
        ypred.to_pickle(save_dir / _default_test_pred_fn)

    return ypred


def score_predictions(test_set: pd.DataFrame, ypred: pd.DataFrame):
    """ Given a test set and predictions made on that test set, returns the Kendall-Tau
    rank correlation (KT) and Coefficient of Determination (R2) scores. """

    _log.info("Extracting the input features and expected labels from the test set.")
    xtest = test_set.loc[:, "features"]
    ytest = test_set.loc[:, "labels"]
    assert xtest.shape[0] == ytest.shape[0]

    _log.info(f"Generating quality of fit scores for {ypred.shape[0]} samples and "
              f"{ypred.shape[1]} outputs.")
    scores = {}
    for output in ytest.columns:
        r2_score = sklearn.metrics.r2_score(ytest[output], ypred[output])

        # No need to adjust for minimization when comparing two series of the same metric.
        ranks_test = ytest[output].rank()
        ranks_pred = ypred[output].rank()
        # kt_corr, kt_p = scipy.stats.kendalltau(ytest[output], ypred[output])
        kt_corr, kt_p = scipy.stats.kendalltau(ranks_test, ranks_pred)
        scores[output] = {
            "R2": float(r2_score),
            "KT": [float(kt_corr), float(kt_p)]
        }

    _log.info(f"Generated scores:\n{scores}")
    return scores


def main(testset_file: Path, model_dir: Optional[Path] = None,
         outputs: Optional[Sequence[str]] = None, scores_only: bool = False,
         predictions_only: bool = False, save_dir: Optional[Path] = None):
    assert testset_file is not None
    test_set = load_test_set(testset_file, outputs)

    if scores_only:
        assert save_dir is not None
        ypred = pd.read_pickle(save_dir / "test_pred.pkl.gz")
    else:
        if predictions_only:
            assert save_dir is not None, \
                "When 'predictions_only' is True, 'save_dir' cannot be None."
        ypred = evaluate_test_set(test_set, model_dir, outputs, save_dir)

    if not predictions_only:
        scores = score_predictions(test_set, ypred)
        if save_dir is not None:
            with open(save_dir / "scores.yaml", "w") as fp:
                yaml.safe_dump(scores)

    _log.info(f"Finished.")


def parse_cli():
    parser = argparse.ArgumentParser(
        "Script to generate scores on a test set for a trained surrogate model."
    )
    parser.add_argument("--testset-file", type=Path, default=None,
                        help="A path to the file which contains the test set.")
    parser.add_argument("--model-dir", type=Path, default=None,
                        help="The directory where the trained surrogate models for each "
                             "relevant output are stored. Optional only when "
                             "--scores-only is given.")
    parser.add_argument("--save-dir", type=Path, default=None,
                        help="An optional directory where the pandas DataFrame of "
                             "predictions made by this script will be stored in a file "
                             "called 'test_pred.pkl.gz' and the scores will be stored in "
                             "a file called 'scores.yaml'. If not provided, the scores "
                             "and predictions are not saved to disk. Must be given if "
                             "--scores-only is given.")
    parser.add_argument("--scores-only", action="store_true", default=False,
                        help="When this flag is given, predictions are not generated "
                             "from a surrogate model. Instead, predictions from a "
                             "previous run are expected to be readable from "
                             "'save-dir/test_pred.pkl.gz'. Scores are generated using "
                             "these predictions. When this flag is given, there is no "
                             "need to provide --model_dir, but --save-dir must be given. "
                             "If this flag is not given, any previously generated "
                             "predictions will get overwritten by newly generated "
                             "predictions.")
    parser.add_argument("--predictions-only", action="store_true", default=False,
                        help="When this flag is given, only predictions are generated "
                             "from a surrogate model, without calculating the "
                             "assosciated scores. When this flag is given, both "
                             "--model_dir and --save-dir must be given. "
                             "If this flag is given, any previously generated "
                             "predictions will get overwritten by newly generated "
                             "predictions.")
    parser.add_argument("--outputs", type=str, default=None,
                        nargs=argparse.REMAINDER,
                        help="Strings, separated by spaces, that indicate which of the "
                             "available metrics in the dataset should be used for "
                             "generating the predictions. If not given, all metrics "
                             "present in the full dataset are used. Note that a model "
                             "for each output is expected to be present in 'save_dir'.")

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = parse_cli()
    logging.basicConfig(level=logging.INFO,
                        format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
                        datefmt="%m/%d %H:%M:%S")
    _log.setLevel(logging.INFO)
    main(**vars(args))



import torch.nn as nn
import ConfigSpace as CS
from typing import Union, Optional

from jahs_bench.tabular.lib.naslib.search_spaces.core.primitives import \
    AbstractPrimitive, Identity
from .constants import Activations


class ConvBN(AbstractPrimitive):

    def __init__(self, C_in, C_out, kernel_size, stride=1, affine=True, activation=Activations.ReLU.name, **kwargs):
        super().__init__(locals())
        self.kernel_size = kernel_size
        pad = 0 if stride == 1 and kernel_size == 1 else 1
        activation_fn = Activations.__members__[activation].value[1]
        self.op = nn.Sequential(
            # nn.SiLU(inplace=False) if use_swish else nn.ReLU(inplace=False),
            activation_fn(inplace=False),
            nn.Conv2d(C_in, C_out, kernel_size, stride=stride, padding=pad, bias=False),
            nn.BatchNorm2d(C_out, affine=affine)
        )


    def forward(self, x, edge_data):
        return self.op(x)

    def get_embedded_ops(self):
        return None

    @property
    def get_op_name(self):
        op_name = super().get_op_name
        op_name += '{}x{}'.format(self.kernel_size, self.kernel_size)
        return op_name


class StemGrayscale(AbstractPrimitive):
    """
    This is used as an initial layer directly after the
    image input in the case of grayscale input images.
    """

    def __init__(self, C_out, **kwargs):
        super().__init__(locals())
        self.seq = nn.Sequential(
            nn.Conv2d(1, C_out, 3, padding=1, bias=False),
            nn.BatchNorm2d(C_out))

    def forward(self, x, edge_data):
        return self.seq(x)

    def get_embedded_ops(self):
        return None


"""
Code below from NASBench-201 and adapted
@inproceedings{dong2020nasbench201,
  title     = {NAS-Bench-201: Extending the Scope of Reproducible Neural Architecture Search},
  author    = {Dong, Xuanyi and Yang, Yi},
  booktitle = {International Conference on Learning Representations (ICLR)},
  url       = {https://openreview.net/forum?id=HJxyZkBKDr},
  year      = {2020}
}
"""


class ResNetBasicblock(AbstractPrimitive):

    def __init__(self, C_in, C_out, stride, affine=True, activation=Activations.ReLU.name):
        super().__init__(locals())
        assert stride == 1 or stride == 2, 'invalid stride {:}'.format(stride)
        self.conv_a = ConvBN(C_in, C_out, 3, stride, activation=activation)
        self.conv_b = ConvBN(C_out, C_out, 3, activation=activation)
        if stride == 2:
            self.downsample = nn.Sequential(
                nn.AvgPool2d(kernel_size=2, stride=2, padding=0),
                nn.Conv2d(C_in, C_out, kernel_size=1, stride=1, padding=0, bias=False))
        else:
            self.downsample = None


    def forward(self, x, edge_data):
        basicblock = self.conv_a(x, None)
        basicblock = self.conv_b(basicblock, None)
        residual = self.downsample(x) if self.downsample is not None else x
        return residual + basicblock


    def get_embedded_ops(self):
        return None
